# pyright: reportArgumentType=false
from itertools import product
from typing import cast, Union

from relationalai.early_access.dsl.core.namespaces import Namespace
from relationalai.early_access.dsl.core.types import Type
from relationalai.early_access.dsl.core.types.constrained.nominal import ValueType
from relationalai.early_access.dsl.ontologies.relationships import Relationship, Reading
from relationalai.early_access.dsl.ontologies.roles import Role
from relationalai.early_access.dsl.bindings.tables import Binding, ColumnRef, SubtypeBinding, IdentifierBinding, \
    RoleBinding, SnowflakeTable
from relationalai.early_access.dsl.codegen.relations import EntityMap, ValueMap, EntitySubtypeMap, RoleMap
from relationalai.early_access.dsl.core.relations import rule, Relation
from relationalai.early_access.dsl.core.rules import Annotation, Vars
from relationalai.early_access.dsl.core.types import AbstractValueType
from relationalai.early_access.dsl.core.types.standard import RowId
from relationalai.early_access.dsl.core.utils import camel_to_snake
from relationalai.early_access.dsl.ontologies.constraints import Unique, Mandatory
from relationalai.early_access.dsl.types.entities import EntityType
from relationalai.early_access.metamodel.util import OrderedSet


class PotentiallyBoundRelationship:
    _relationship: 'Relationship'
    _table: 'SnowflakeTable'
    _bindings: list['Binding']

    def __init__(self, relationship: 'Relationship', table: 'SnowflakeTable'):
        self._relationship = relationship
        self._table = table
        self._bindings = []

    def __hash__(self):
        # hash based on the relationship and the table
        return hash((self._relationship, self._table))

    @property
    def relationship(self):
        return self._relationship

    @property
    def table(self):
        return self._table

    @property
    def bindings(self):
        return self._bindings


class Weaver:
    # Bindings
    _value_type_bindings: OrderedSet['Binding'] = OrderedSet()
    _entity_type_bindings: OrderedSet['Binding'] = OrderedSet()
    _constructor_bindings: OrderedSet[Union['IdentifierBinding', 'SubtypeBinding']] = OrderedSet()
    _referent_constructor_bindings: OrderedSet[Union['IdentifierBinding', 'SubtypeBinding']] = OrderedSet()
    _referent_bindings: OrderedSet['RoleBinding'] = OrderedSet()
    _subtype_bindings: OrderedSet['SubtypeBinding'] = OrderedSet()
    _subtype_binding_references: dict['SubtypeBinding', 'Role'] = {}
    _constructor_binding_binds_to: dict['IdentifierBinding', 'Role'] = {}
    _role_bound_thru: dict['Role', list['Binding']] = {}
    _ref_binding_to_ctor_binding: dict['Binding', 'Binding'] = {}

    # Relationships & Ontological Constraints
    _unary_relationships: OrderedSet['Relationship'] = OrderedSet()
    _binary_relationships: OrderedSet['Relationship'] = OrderedSet()
    _identifier_relationships: OrderedSet['Relationship'] = OrderedSet()
    _entity_identifier_relationships: OrderedSet['Relationship'] = OrderedSet()
    _entity_type_identifier: dict['EntityType', 'Relationship'] = {}
    _bound_relationships: OrderedSet['Relationship'] = OrderedSet()
    _preferred_uc: list['Unique'] = []
    _constructor_roles: OrderedSet['Role'] = OrderedSet()
    _internal_uc_roles: OrderedSet['Role'] = OrderedSet()
    _mandatory_roles: OrderedSet['Role'] = OrderedSet()
    _bound_role: dict['Role', list['Binding']] = {}
    _subtype_of: dict['EntityType', 'EntityType'] = {}
    _inclusive_entity_types: OrderedSet['EntityType'] = OrderedSet()
    _composite_entity_types: OrderedSet['EntityType'] = OrderedSet()
    _subtype_map: dict['EntityType', OrderedSet['EntityType']] = {}
    _subtype_closure: dict['EntityType', OrderedSet['EntityType']] = {}
    _supertype_map: dict['EntityType', 'EntityType'] = {}

    # Physical Relations & Classification
    _relations: dict[str, 'Relation'] = {}
    _preferred_id_relations: OrderedSet['Relation'] = OrderedSet()
    _identifier_relations: OrderedSet['Relation'] = OrderedSet()
    _entity_map_relations: OrderedSet['Relation'] = OrderedSet()
    _subtype_entity_map_relations: OrderedSet['Relation'] = OrderedSet()
    _value_map_relations: OrderedSet['Relation'] = OrderedSet()
    _entity_population_relations: OrderedSet['Relation'] = OrderedSet()

    _binding_to_value_map: dict['Binding', 'Relation'] = {}
    _constructor_binding_to_value_map: dict['Binding', 'Relation'] = {}
    _col_ref_to_pid_concept: dict['ColumnRef', 'EntityType'] = {}
    _entity_to_id_relations: dict['EntityType', 'Relation'] = {}
    _concept_to_identifies_relations: dict['Type', 'Relation'] = {}
    _subtype_to_entity_map: dict['EntityType', 'Relation'] = {}

    def __init__(self, model):
        self._model = model

    def analyze(self):
        self._analyze_subtypes()
        potentially_bound_rels = self._analyze_bindings()
        self._analyze_constraints()
        self._analyze_relationships()
        self._analyze_bound_relationships(potentially_bound_rels)

    def _analyze_subtypes(self):
        for subtype_arrow in self._model._subtype_arrows:
            parent = subtype_arrow.end
            if parent not in self._subtype_map:
                self._subtype_map[parent] = OrderedSet()
            self._subtype_map[parent].add(subtype_arrow.start)
            self._supertype_map[subtype_arrow.start] = parent

        self._subtype_closure = {parent: OrderedSet() for parent in self._subtype_map}
        for parent in self._subtype_map.keys():
            self._subtype_closure_dfs(parent, parent)

    def _subtype_closure_dfs(self, prt, curr):
        for chld in self._subtype_map.get(curr, []):
            cls = self._subtype_closure[prt]
            if chld not in cls:
                cls.add(chld)
                self._subtype_closure_dfs(prt, chld)

    def _analyze_bindings(self):
        potentially_bound_rels = {}
        for binding in self._model._bindings:
            role = self._process_binding(binding)
            assert role is not None
            if role not in self._bound_role:
                self._bound_role[role] = []
            self._bound_role[role].append(binding)
            self._categorize_binding(binding, role)

            # Group bindings by relationship and table
            rel = PotentiallyBoundRelationship(role.part_of, binding.column.table)
            if rel not in potentially_bound_rels:
                potentially_bound_rels[rel] = []
            potentially_bound_rels[rel].append(binding)
        return potentially_bound_rels

    def _process_binding(self, binding: 'Binding'):
        if isinstance(binding, SubtypeBinding):
            binding = cast(SubtypeBinding, binding)
            role = self._lookup_ctor_role(binding.parent_type)
            self._subtype_binding_references[binding] = role
            self._subtype_bindings.add(binding)
            self._subtype_of[binding.sub_type] = binding.parent_type
        elif isinstance(binding, IdentifierBinding):
            binding = cast(IdentifierBinding, binding)
            role = self._lookup_ctor_role(binding.entity_type)
            self._constructor_binding_binds_to[binding] = role
        elif isinstance(binding, RoleBinding):
            binding = cast(RoleBinding, binding)
            role = binding.role
        else:
            raise Exception(f'Binding {binding} is not supported')
        return role

    def _categorize_binding(self, binding: 'Binding', role: 'Role'):
        player = role.player()
        if isinstance(player, AbstractValueType):
            self._value_type_bindings.add(binding)
        elif isinstance(player, EntityType):
            self._entity_type_bindings.add(binding)
        else:
            raise Exception(f'Binding {binding} is not supported')

    def _analyze_constraints(self):
        for constraint in self._model.constraints():
            if isinstance(constraint, Unique):
                self._process_unique_constraint(constraint)
            elif isinstance(constraint, Mandatory):
                self._mandatory_roles.update(constraint.roles())

    def _process_unique_constraint(self, constraint: 'Unique'):
        if not constraint.is_preferred_identifier:
            return

        self._preferred_uc.append(constraint)
        roles = constraint.roles()
        self._constructor_roles.update(roles)  # Mark as constructor roles

        if len(roles) != 1:
            raise Exception('Complex preferred identifier case not supported yet')

        constructor_role = roles[0]
        rel = constructor_role.part_of

        if rel.arity() != 2:
            raise Exception('Invalid Identifier relationship configuration')

        player = constructor_role.player()
        if isinstance(player, AbstractValueType):
            self._identifier_relationships.add(rel)
        elif isinstance(player, EntityType):
            self._entity_identifier_relationships.add(rel)
        else:
            raise Exception(f'Identifier relationship {rel.pprint()} has unsupported player type {player.name()}')

        concept = constructor_role.sibling().player()
        self._entity_type_identifier[concept] = rel
        self._inclusive_entity_types.add(concept)

        # Classify bindings
        if constructor_role not in self._bound_role:
            return

        for binding in self._bound_role[constructor_role]:
            if isinstance(binding, IdentifierBinding):
                if not binding.column.references:
                    role = self._lookup_binding_role(binding)
                    self._col_ref_to_pid_concept[binding.column.ref()] = role.player()
                    self._constructor_bindings.add(binding)
                else:
                    self._referent_constructor_bindings.add(binding)
            elif isinstance(binding, SubtypeBinding):
                self._referent_constructor_bindings.add(binding)

    def _analyze_relationships(self):
        for relationship in self._model.relationships():
            arity = relationship.arity()
            if arity == 1:
                self._unary_relationships.add(relationship)
            elif arity == 2:
                self._binary_relationships.add(relationship)

    def _analyze_bound_relationships(self, potentially_bound_rels):
        for rel, bindings in potentially_bound_rels.items():
            rel.bindings.extend(bindings)
            self._analyze_bound_relationship(rel)

    def _lookup_ctor_role(self, entity_type: 'EntityType'):
        # TODO: implement bottom up lookup
        id_rel = self._model.identifier_of(entity_type)
        if id_rel is None:
            raise Exception(f'Identifier relationship for {entity_type} not found')
        for role in id_rel.roles():
            if role.player() == entity_type:
                # ctor role is the only sibling of the constructed entity type role
                return role.sibling()

    def _analyze_bound_relationship(self, meta: 'PotentiallyBoundRelationship'):
        rel = meta.relationship
        arity = rel.arity()

        if arity == 1:
            # Handle unary relationships
            if meta.bindings:
                self._bound_relationships.add(meta)
            return

        if arity != 2:
            raise Exception('N-ary (3+) bound relationship case not supported yet')

        # Handle binary relationships
        key_role, value_role = self._identify_roles(rel)
        key_type, value_type = key_role.player(), value_role.player()

        # TODO : this below should really check if they bound for the same table
        key_role_is_bound = key_role in self._bound_role
        value_role_is_bound = value_role in self._bound_role
        value_role_is_ctor = value_role in self._constructor_roles

        key_role_infers_emap = self._infer_key_role_emap(key_type, meta.table, key_role_is_bound, value_role_is_ctor)
        fully_bound = value_role_is_bound and (key_role_is_bound or key_role_infers_emap)

        is_bound_value_type_relationship = isinstance(value_type, AbstractValueType) and fully_bound
        is_bound = self._is_bound_identifier_relationship(value_role_is_ctor, value_role_is_bound, key_role_is_bound) or \
                   is_bound_value_type_relationship or \
                   self._is_bound_entity_type_relationship(key_type, value_type, key_role, value_role, meta.table)
        if is_bound:
            self._bound_relationships.add(meta)
        else:
            raise Exception('Bound relationship must have at least one bound role and one inferred entity map')

    def _identify_roles(self, rel: 'Relationship') -> tuple['Role', 'Role']:
        key_role, value_role = None, None
        for role in rel.roles():
            if isinstance(role.player(), AbstractValueType):
                value_role = role
            # if both roles are entity types, arbitrarily pick one as the key role
            elif key_role is None and role not in self._constructor_roles:
                key_role = role
            else:
                value_role = role
        assert key_role is not None
        assert value_role is not None
        return key_role, value_role

    def _infer_key_role_emap(self, key_type, table, key_role_is_bound, value_role_is_ctor):
        if key_role_is_bound or value_role_is_ctor:
            return False
        return self._exists_ctor_binding(key_type, table) or self._exists_ref_binding(key_type, table)

    @staticmethod
    def _is_bound_identifier_relationship(value_role_is_ctor, value_role_is_bound, key_role_is_bound):
        if value_role_is_ctor and value_role_is_bound:
            if key_role_is_bound:
                raise Exception('Identifier relationship must not have the key role bound')
            return True
        return False

    def _is_bound_entity_type_relationship(self, key_type, value_type, key_role, value_role, table):
        if isinstance(value_type, EntityType) and isinstance(key_type, EntityType):
            value_role_infers_emap = self._exists_ctor_binding(value_type, table) or self._exists_ref_binding(value_type, table)
            return (
                (key_role in self._bound_role and value_role in self._bound_role) or
                (value_role in self._bound_role and value_role_infers_emap) or
                (key_role in self._bound_role and value_role_infers_emap)
            )
        if isinstance(value_type, AbstractValueType) and isinstance(key_type, AbstractValueType):
            raise Exception('Binary Relationship cannot have more than one ValueType role')
        return False

    def _exists_ctor_binding(self, entity_type: 'EntityType', table: 'SnowflakeTable'):
        return self._lookup_ctor_binding(entity_type, table) is not None

    def _lookup_ctor_binding(self, entity_type: 'EntityType', table: 'SnowflakeTable'):
        # TODO : implement more efficient lookup
        for binding, role in self._constructor_binding_binds_to.items():
            sibling_role = role.sibling()
            assert sibling_role is not None
            if sibling_role.player() == entity_type and binding.column.table == table:
                return binding
        if entity_type in self._subtype_closure:
            for subtype in self._subtype_closure[entity_type]:
                cand = self._lookup_ctor_binding(subtype, table)
                if cand:
                    return cand
        return None

    def _exists_ref_binding(self, entity_type: 'EntityType', table: 'SnowflakeTable'):
        return self._lookup_ref_binding(entity_type, table) is not None

    def _lookup_ref_binding(self, entity_type: 'EntityType', table: 'SnowflakeTable'):
        # TODO : implement more efficient lookup
        for binding, role in self._subtype_binding_references.items():
            if isinstance(binding, SubtypeBinding):
                binding = cast(SubtypeBinding, binding)
                sibling_role = role.sibling()
                assert sibling_role is not None
                if sibling_role.player() == binding.parent_type and binding.column.table == table and entity_type == binding.sub_type:
                    return binding
        if entity_type in self._subtype_closure:
            for subtype in self._subtype_closure[entity_type]:
                cand = self._lookup_ref_binding(subtype, table)
                if cand:
                    return cand
        return None

    def _lookup_binding_role(self, binding: 'Binding'):
        if isinstance(binding, IdentifierBinding):
            role = self._constructor_binding_binds_to[binding]
        elif isinstance(binding, RoleBinding):
            role = binding.role
        elif isinstance(binding, SubtypeBinding):
            role = self._subtype_binding_references[binding]
        else:
            raise Exception(f'Binding {binding} is not supported')
        return role

    def generate(self):
        self._process_bound_relationships()
        self._process_referent_bindings()
        self._generate_value_maps()
        self._generate_entity_maps()
        self._generate_subtype_entity_maps()
        self._generate_entity_populations()
        self._generate_non_identifier_relationships()

    def _process_bound_relationships(self):
        for rel_meta in self._bound_relationships:
            if self._is_identifier_relationship(rel_meta.relationship):
                self._process_identifier_relationship(rel_meta)
            else:
                self._process_non_identifier_relationship(rel_meta)

    def _is_identifier_relationship(self, relationship):
        return relationship in self._identifier_relationships or \
               relationship in self._entity_identifier_relationships

    def _process_identifier_relationship(self, rel_meta):
        roles = rel_meta.relationship.roles()
        ctor_role, concept_role = self._classify_identifier_roles(roles)
        ctor_role_bindings, concept_role_bindings = self._categorize_bindings(rel_meta.bindings)
        self._update_role_bound_thru(ctor_role, ctor_role_bindings)
        self._update_role_bound_thru(concept_role, concept_role_bindings)

    def _classify_identifier_roles(self, roles):
        ctor_role = next((role for role in roles if role in self._constructor_roles), None)
        concept_role = next((role for role in roles if role != ctor_role), None)
        return ctor_role, concept_role

    def _categorize_bindings(self, bindings):
        ctor_role_bindings = []
        concept_role_bindings = []
        for binding in bindings:
            if binding in self._constructor_bindings:
                ctor_role_bindings.append(binding)
                concept_role_bindings.append(binding)
            elif binding in self._referent_constructor_bindings:
                ctor_role_bindings.append(binding)
                concept_role_bindings.append(binding)
            elif binding in self._subtype_bindings:
                concept_role_bindings.append(binding)
        return ctor_role_bindings, concept_role_bindings

    def _update_role_bound_thru(self, role, bindings):
        if role not in self._role_bound_thru:
            self._role_bound_thru[role] = []
        self._role_bound_thru[role].extend(bindings)

    def _process_non_identifier_relationship(self, rel_meta):
        role_bindings = {}
        for binding in rel_meta.bindings:
            role = self._lookup_binding_role(binding)
            role_bindings.setdefault(role, []).append(binding)
            if isinstance(binding, RoleBinding) and isinstance(binding.role.player(), EntityType):
                self._referent_bindings.add(binding)
        self._role_bound_thru.update(role_bindings)

    def _process_referent_bindings(self):
        for binding in self._referent_bindings:
            ctor_binding = self._lookup_binding_reference(binding)
            self._ref_binding_to_ctor_binding[binding] = ctor_binding

    def _generate_value_maps(self):
        for binding in self._value_type_bindings:
            self._gen_value_map(binding)

    def _generate_entity_maps(self):
        for binding in self._constructor_bindings:
            self._gen_entity_map(binding)
            self._gen_preferred_id(binding)
        for binding in self._referent_constructor_bindings:
            self._gen_entity_map(binding)

    def _generate_subtype_entity_maps(self):
        for binding in self._subtype_bindings:
            self._gen_subtype_entity_map(binding)

    def _generate_entity_populations(self):
        for entity_type, relation in self._entity_to_id_relations.items():
            self._gen_entity_population(entity_type, relation)
        for subtype, relation in self._subtype_to_entity_map.items():
            self._gen_entity_population(subtype, relation)

    def _generate_non_identifier_relationships(self):
        for rel_meta in self._bound_relationships:
            if not self._is_identifier_relationship(rel_meta.relationship):
                self._gen_semantic_predicate(rel_meta)

    def _lookup_binding_reference(self, binding):
        ref_concept = binding.role.player()
        # case 1: ref_concept is an entity type with a reference scheme - just look up the ID relationship
        if ref_concept in self._inclusive_entity_types:
            return self._lookup_inclusive_type_binding(ref_concept, binding)
        # case 2: ref_concept is an entity type with a composite reference scheme
        elif ref_concept in self._composite_entity_types:
            # TODO implement composite reference scheme
            raise Exception('Composite reference scheme not supported yet')
        # case 3: ref_concept has no reference scheme, i.e. exclusive entity type
        else:
            # In this case we look up an IdentifierBinding that has the same column as the binding
            # and role player as the same entity type played by the binding. It must be unique,
            # otherwise we raise an exception.
            #
            # NOTE: this is oversimplified now, as we check subtype closure AND immediate supertype.
            cand_bindings = []
            supertype = self._supertype_map.get(ref_concept)
            if supertype and supertype in self._inclusive_entity_types:
                cand = self._lookup_inclusive_type_binding(supertype, binding)
                if cand:
                    cand_bindings.append(cand)
            else:
                subtype_closure = self._subtype_closure[ref_concept]
                for cand_binding in self._constructor_bindings:
                    ctor_role = self._lookup_binding_role(cand_binding)
                    # to be precise, we need to compare with the sibling role, as it is the one that
                    # is played by the entity type in question
                    sibling_role = ctor_role.sibling()
                    assert sibling_role is not None
                    concept = sibling_role.player()
                    assert isinstance(concept, EntityType)
                    if cand_binding.column == binding.column and concept in subtype_closure:
                        cand_bindings.append(cand_binding)
            if len(cand_bindings) != 1:
                raise Exception(f'Binding {binding} has no unique reference to an IdentifierBinding')
            else:
                return cand_bindings[0]

    def _lookup_inclusive_type_binding(self, ref_concept: 'EntityType', binding: 'Binding'):
        identifier_relationship = self._model.identifier_of(ref_concept)
        # TODO optimize
        ctor_binding = None
        concept_role = None
        for role in identifier_relationship.roles():
            if role in self._constructor_roles:
                concept_role = role.sibling()
            else:
                concept_role = role
            break
        for cand_binding in self._role_bound_thru[concept_role]:
            if isinstance(cand_binding, IdentifierBinding) and binding.column == cand_binding.column:
                ctor_binding = cand_binding
                break
        return ctor_binding

    def _gen_preferred_id(self, binding: Union['IdentifierBinding', 'SubtypeBinding']):
        # =
        # Simple case: single role, hence the preferred id is the role itself.
        #
        # We generate the following rules:
        #
        #   def {concept}:id(v, c): ...
        # =
        ctor_role = self._lookup_binding_role(binding)
        value_concept = ctor_role.player()
        val_role = ctor_role.sibling()
        assert val_role is not None
        entity_concept = val_role.player()

        # find the relations that have been created
        pref_id_relation = None
        identifies_relation = None
        for reading in ctor_role.part_of.readings():
            roles = reading.roles
            rel_name = reading.rel_name
            if roles[0] == ctor_role:
                if isinstance(value_concept, EntityType) or isinstance(value_concept, ValueType):
                    identifies_relation = getattr(value_concept, rel_name)
            else:
                assert isinstance(entity_concept, EntityType)
                pref_id_relation = entity_concept[rel_name]

        rel_nm = self._fqn_preferred_id(entity_concept)
        assert pref_id_relation is not None
        with pref_id_relation:
            @rule()
            def identifier(c, i):
                row = Vars(RowId)
                self._ref_value_map(binding)(row, i)
                self._ref_constructor_entity_map(binding)(row, c)
        self._register_relation(rel_nm, pref_id_relation, self._preferred_id_relations)
        self._entity_to_id_relations[entity_concept] = pref_id_relation

        # multiple bindings can exist, but we only need to transpose once
        rel_nm = self._fqn_identifies(value_concept)
        # TODO: fix this one below
        if isinstance(value_concept, AbstractValueType) and rel_nm not in self._relations:
            identifies_relation = self._model.external_relation(rel_nm, value_concept, entity_concept)
        assert identifies_relation is not None
        if identifies_relation not in self._identifier_relations:
            assert pref_id_relation is not None
            with identifies_relation:
                @rule()
                def identifies(i, c):
                    pref_id_relation(c, i)
            self._register_relation(rel_nm, identifies_relation, self._identifier_relations)
            self._concept_to_identifies_relations[value_concept] = identifies_relation

    @staticmethod
    def _fqn_preferred_id(concept: 'EntityType'):
        return f'{camel_to_snake(concept.name())}__id'

    def _ref_identifies(self, binding: 'Binding'):
        pid_ref = binding.column.references
        if pid_ref:
            concept = self._col_ref_to_pid_concept[pid_ref]
        else:
            role = self._lookup_binding_role(binding)
            concept = role.player()
        fqn = self._fqn_identifies(concept)
        return self._lookup_relation_by_fqn(fqn)

    @staticmethod
    def _fqn_identifies(concept: 'Type'):
        if isinstance(concept, EntityType):
            concept = cast(EntityType, concept)
            ref_schema_nm = concept.ref_schema_name()
        else:
            ref_schema_nm = 'identifies'
        return f'{camel_to_snake(concept.name())}__{ref_schema_nm}'

    def _transpose(self, relation: 'Relation', fqn: str):
        if fqn in self._relations:
            return self._relations[fqn]
        if relation.arity() != 2:
            raise Exception('Transposition only supported for binary relations')
        (left_concept, right_concept) = relation.signature().types()
        rel = self._model.external_relation(fqn, right_concept, left_concept)
        with rel:
            @rule()
            def transpose(left, right):
                relation(right, left)
        self._relations[fqn] = rel
        return rel

    def _gen_entity_map(self, binding: 'Binding'):
        vt_role = self._lookup_binding_role(binding)
        et_role = vt_role.sibling()
        assert et_role is not None
        value_concept = vt_role.player()
        entity_concept = et_role.player()
        assert isinstance(entity_concept, EntityType)
        rel_nm = self._fqn_entity_map_indexed(binding, et_role)
        rel = EntityMap(Namespace.top, rel_nm, RowId, binding.column.relation(), et_role)
        role_map = self._ref_value_map(binding) if isinstance(value_concept,
                                                              AbstractValueType) else self._ref_constructor_entity_map(
            binding)
        with rel:
            @rule()
            def entity_map(row, et):
                val = Vars(value_concept)
                role_map(row, val)
                if binding.column.references:
                    self._ref_identifies(binding)(val, et)
                else:
                    _ = entity_concept ^ (val, et)  # `_=` is not strictly needed, but it makes IDEs happy ;)
        self._register_relation(rel_nm, rel, self._entity_map_relations)
        self._constructor_binding_to_value_map[binding] = rel

    def _ref_constructor_entity_map(self, binding: 'Binding'):
        role = self._lookup_binding_role(binding)
        return self._ref_entity_map(binding, role.sibling())

    def _fqn_entity_map_indexed(self, binding: 'Binding', role: 'Role'):
        role_bindings = self._role_bound_thru[role]
        idx = role_bindings.index(binding)
        fqn = self._fqn_entity_map(binding, role.player())
        fqn = f'{fqn}{idx}' if idx > 0 else fqn
        return fqn

    def _ref_entity_map(self, binding: 'Binding', role: 'Role'):
        fqn = self._fqn_entity_map_indexed(binding, role)
        return self._lookup_relation_by_fqn(fqn)

    def _ref_subtype_entity_map(self, binding: 'SubtypeBinding'):
        fqn = self._fqn_entity_map(binding, binding.sub_type)
        return self._lookup_relation_by_fqn(fqn)

    @staticmethod
    def _fqn_entity_map(binding: 'Binding', concept: 'EntityType'):
        assert isinstance(concept, EntityType)
        concept_nm = camel_to_snake(concept.name())
        source_nm = binding.column.table.physical_name()
        return f'{concept_nm}__impl__{source_nm}_row_to_{concept_nm}'

    def _gen_subtype_entity_map(self, binding: 'SubtypeBinding'):
        subtype = binding.sub_type
        rel_nm = self._fqn_entity_map(binding, subtype)
        rel = EntitySubtypeMap(Namespace.top, rel_nm, RowId, binding)
        with rel:
            @rule()
            def subtype_entity_map(row, et):
                self._ref_constructor_entity_map(binding)(row, et)
        self._register_relation(rel_nm, rel, self._subtype_entity_map_relations)
        self._subtype_to_entity_map[subtype] = rel

    def _gen_value_map(self, binding: 'Binding'):
        role = self._lookup_binding_role(binding)
        rel_nm = self._fqn_value_map_indexed(binding)
        rel = ValueMap(Namespace.top, rel_nm, RowId, binding.column.relation(), role)
        with rel:
            @rule(Annotation.INLINE)
            def value_map(row, val):
                rel.attr_view()(row, val)
        self._binding_to_value_map[binding] = rel
        self._register_relation(rel_nm, rel, self._value_map_relations)

    def _fqn_value_map(self, binding: 'Binding'):
        role = self._lookup_binding_role(binding)
        concept = role.player()
        assert isinstance(concept, AbstractValueType)
        concept_nm = camel_to_snake(concept.name())
        source_nm = binding.column.table.physical_name()
        return f'{concept_nm}__impl__{source_nm}_row_to_{concept_nm}'

    def _fqn_value_map_indexed(self, binding: 'Binding'):
        role = self._lookup_binding_role(binding)
        role_bindings = self._role_bound_thru[role]
        idx = role_bindings.index(binding)
        fqn = self._fqn_value_map(binding)
        fqn = f'{fqn}{idx}' if idx > 0 else fqn
        return fqn

    def _ref_value_map(self, binding: 'Binding'):
        fqn = self._fqn_value_map_indexed(binding)
        return self._lookup_relation_by_fqn(fqn)

    def _gen_entity_population(self, entity_type: 'EntityType', lookup_rel: 'Relation'):
        rel_nm = self._fqn_entity_population(entity_type)
        rel = self._model.external_relation(rel_nm, entity_type)
        if lookup_rel in self._preferred_id_relations:
            val_type = self._get_last_type(lookup_rel)
            with rel:
                @rule()
                def entity_population(et):
                    row = Vars(val_type)
                    lookup_rel(et, row)
        elif lookup_rel in self._subtype_entity_map_relations:
            with rel:
                @rule()
                def entity_population(et):
                    row = Vars(RowId)
                    lookup_rel(row, et)
        else:
            raise Exception(f'Unsupported weaving type for relation `{lookup_rel.qualified_name()}`')
        self._register_relation(rel_nm, rel, self._entity_population_relations)

    @staticmethod
    def _fqn_entity_population(entity_type: 'EntityType'):
        # entity population name is the same as the entity type name, with the first letter uppercased
        return entity_type.name()

    def _lookup_relation_by_fqn(self, fqn: str):
        if fqn in self._relations:
            return self._relations[fqn]
        else:
            raise Exception(f'Relation with fully qualified name {fqn} not found in the model')

    def _gen_semantic_predicate(self, rel_meta: 'PotentiallyBoundRelationship'):
        relationship = rel_meta.relationship
        if relationship.arity() > 2:
            raise Exception(
                f'Cannot generate semantic predicate for relationship {relationship.pprint()} with arity {relationship.arity()} (more than 2)')
        roles = relationship.roles()
        # Each role should either be covered by a value map (AbstractValueType role) or by an entity map
        # (EntityType role), which either can be inferred if unique exists for the EntityType or must be
        # generated by the respective EntityBinding.
        role_to_role_map = {}
        for role in roles:
            player = role.player()
            # for ValueTypes just look up ValueMap relations
            if isinstance(player, AbstractValueType):
                if role not in self._bound_role:
                    raise Exception(f'ValueType role {role.name()} is not bound')
                else:
                    value_maps = [self._ref_value_map(binding) for binding in self._bound_role[role]]
                    if len(value_maps) == 0:
                        raise Exception(f'ValueType role {role.name()} is not correctly bound')
                    if role not in role_to_role_map:
                        role_to_role_map[role] = OrderedSet()
                    role_to_role_map[role].update(value_maps)
            elif isinstance(player, EntityType):
                entity_maps = OrderedSet()
                if role not in self._bound_role:
                    if player in self._subtype_of:
                        binding = self._lookup_ref_binding(player, rel_meta.table)
                        if binding is None:
                            raise Exception(f'EntityType({player.name()}) role {role.name()} is not bound')
                        entity_map = self._ref_subtype_entity_map(binding)
                        entity_maps.add(entity_map)
                    else:
                        binding = self._lookup_ctor_binding(player, rel_meta.table)
                        if binding is None:
                            raise Exception(f'EntityType({player.name()}) role {role.name()} is not bound')
                        entity_map = self._ref_constructor_entity_map(binding)
                        entity_maps.add(entity_map)
                else:
                    bindings = self._bound_role[role]
                    for binding in bindings:
                        if isinstance(binding, RoleBinding) and binding in self._referent_bindings:
                            binding = self._ref_binding_to_ctor_binding[binding]
                        entity_map = self._ref_constructor_entity_map(binding)
                        assert entity_map is not None
                        entity_maps.add(entity_map)
                if len(entity_maps) == 0:
                    raise Exception(f'EntityType role {role.name()} is not correctly bound')
                if role not in role_to_role_map:
                    role_to_role_map[role] = OrderedSet()
                role_to_role_map[role].update(entity_maps)
            else:
                raise Exception(f'Role {role.name()} is not bound to a ValueType or EntityType')
        # if we got all roles with role maps, we can generate the rule
        items = role_to_role_map.items()
        if len(items) != relationship.arity():
            raise Exception(f'Not all roles of relationship {relationship.pprint()} are bound')
        for reading in relationship.readings():
            concept = reading.roles[0].player()
            rel_nm = reading.rel_name
            relation = concept[rel_nm]

            role_map_combinations = self._permute_role_maps(role_to_role_map, as_in=reading)
            for role_map1, role_map2 in role_map_combinations:
                with relation:
                    @rule()
                    def semantic_predicate(c, v):
                        row = Vars(RowId)
                        role_map1(row, c)
                        role_map2(row, v)

    @staticmethod
    def _permute_role_maps(role_to_role_map: dict['Role', OrderedSet['RoleMap']], as_in: 'Reading'):
        try:
            role_map_combinations = product(*(role_to_role_map[role] for role in as_in.roles))
            return [list(combination) for combination in role_map_combinations]
        except KeyError as e:
            missing_role = e.args[0]
            raise Exception(f'Cannot permute role maps to match reading `{as_in.rel_name}`, role `{missing_role.name()}` not found')

    @staticmethod
    def _get_last_type(lookup_rel: 'Relation'):
        return lookup_rel.signature().types()[-1]

    def _register_relation(self, name: str, rel: 'Relation', population: OrderedSet['Relation'] = None):
        if isinstance(rel, RoleMap) or isinstance(rel, EntitySubtypeMap):
            self._model._add_relation(rel)
        self._relations[name] = rel
        if population is not None:
            population.add(rel)



from pandas import DataFrame
import rich
from rich.table import Table
from rich import box
from rich.console import Console
from relationalai.environments.base import SourceInfo
from relationalai.metamodel import Namer
from relationalai.std import rel
from relationalai.environments import runtime_env
from relationalai.dsl import to_var

_watch_info = {}

def watch(*args):
    ix = len(_watch_info)
    namer = Namer()
    _watch_info[ix] = {"loc": runtime_env.get_source(), "args": [namer.get(to_var(a)) for a in args]}
    rel.output.__pyrel_debug_watch.add(ix, *args)

def _watch_table(df:DataFrame):
    # Create a table
    table = Table(show_header=True, box=box.ROUNDED, padding=(0, 1))
    for k in df.columns:
        table.add_column(k)

    for row in df.itertuples(index=False):
        table.add_row(*[str(p) for p in row])

    console = Console()
    width = console.measure(table).maximum
    rich.print(table)
    row_str = f"{len(df)} rows"
    rich.print(f"[dim white]{row_str:>{width-1}}")

def _print_watch_frame(result_frame:DataFrame):
    first_col = result_frame.columns[0]
    for ix, group in result_frame.groupby(first_col):
        source_info:SourceInfo = _watch_info[ix]["loc"]
        line = source_info.source.split("\n")[source_info.line - source_info.source_start_line]
        print()
        rich.print(f"[yellow bold]{line.strip()}")
        rich.print(f"[dim white]{source_info.file}:{source_info.line}")
        print()
        group = group.drop(first_col, axis=1)
        group.columns = _watch_info[ix]["args"]
        group.reset_index(drop=True, inplace=True)
        _watch_table(group)



import dataclasses
from typing import Optional

from relationalai.early_access.dsl.bindings.common import BindableAttribute, BindableTable
from relationalai.early_access.dsl.bindings.relations import AttributeView
from relationalai.early_access.dsl.core.types import Type
from relationalai.early_access.dsl.core.types.standard import String, Integer, Decimal, Date, DateTime, Boolean, \
    BigInteger, Float, RowId
from relationalai.early_access.dsl.core.utils import generate_stable_uuid
from relationalai.early_access.dsl.core.relations import Relation
from relationalai.early_access.dsl.core.namespaces import Namespace
from relationalai.early_access.dsl.types.entities import EntityType
from relationalai.early_access.dsl.ontologies.roles import Role
#=
# Physical metadata for a Snowflake table.
#=

@dataclasses.dataclass(frozen=True)
class ColumnRef:
    table: str
    column: str

@dataclasses.dataclass
class ForeignKey:
    name: str
    source_columns: list[ColumnRef] = dataclasses.field(default_factory=list)
    target_columns: list[ColumnRef] = dataclasses.field(default_factory=list)

    def __hash__(self):
        return hash(self.name)

@dataclasses.dataclass
class SchemaMetadata:
    name: str
    foreign_keys: list[ForeignKey] = dataclasses.field(default_factory=list)

@dataclasses.dataclass
class ColumnMetadata:
    name: str
    datatype: str
    is_nullable: bool
    numeric_precision: Optional[int] = None
    numeric_precision_radix: Optional[int] = None
    numeric_scale: Optional[int] = None

@dataclasses.dataclass
class TabularMetadata:
    name: str
    columns: list[ColumnMetadata] = dataclasses.field(default_factory=list)
    foreign_keys: set[ForeignKey] = dataclasses.field(default_factory=set)

@dataclasses.dataclass(frozen=True)
class Binding:
    column: 'BindableColumn'

@dataclasses.dataclass(frozen=True)
class RoleBinding(Binding):
    role: 'Role'

@dataclasses.dataclass(frozen=True)
class IdentifierBinding(Binding):
    entity_type: 'EntityType'

@dataclasses.dataclass(frozen=True)
class SubtypeBinding(Binding):
    sub_type: 'EntityType'
    parent_type: 'EntityType'

_sf_type_mapping = {
    'varchar': String,
    'char': String,
    'text': String,
    'date': Date,
    'datetime': DateTime,
    'timestamp_ntz': DateTime,
    'boolean': Boolean,
    'float': Float,
}
def _map_rai_type(col: ColumnMetadata) -> 'Type':
    datatype = col.datatype.lower()
    # TODO: do better with integer type conversion
    if datatype == 'number':
        if col.numeric_scale is not None and col.numeric_scale > 0:
            return Decimal
        elif col.numeric_precision is not None and col.numeric_precision > 9:
            return BigInteger
        else:
            return Integer
    else:
        return _sf_type_mapping[datatype]

class BindableColumn(BindableAttribute):
    _metadata: 'ColumnMetadata'
    _table: 'SnowflakeTable'
    _references: Optional['ColumnRef']

    def __init__(self, metadata: 'ColumnMetadata', table: 'SnowflakeTable', model):
        self._metadata = metadata
        self._table = table
        self._model = model
        self._datatype = _map_rai_type(self._metadata)
        self._references = None
        self._attr_relation = AttributeView(Namespace.top, self)

    def __call__(self, *args):
        return self.relation()(*args)

    def identifies(self, entity_type: 'EntityType'):
        binding = IdentifierBinding(column=self, entity_type=entity_type)
        self._model.role_binding(binding)

    def binds_to(self, role: 'Role'):
        binding = RoleBinding(role=role, column=self)
        self._model.role_binding(binding)

    def references_subtype(self, sub_type: 'EntityType', parent_type: 'EntityType'):
        binding = SubtypeBinding(column=self, sub_type=sub_type, parent_type=parent_type)
        self._model.role_binding(binding)

    def binds(self, relation: 'Relation'):
        roles = relation.reading().roles
        # this binds to the last role in binary relations
        role = roles[-1]
        binding = RoleBinding(role=role, column=self)
        self._model.role_binding(binding)

    def relation(self):
        return self._attr_relation

    @property
    def metadata(self):
        return self._metadata

    @property
    def table(self):
        return self._table

    @property
    def references(self):
        return self._references

    @references.setter
    def references(self, ref: 'ColumnRef'):
        self._references = ref

    def physical_name(self):
        return self._metadata.name

    def type(self):
        return _map_rai_type(self._metadata)

    def guid(self):
        return generate_stable_uuid(f'$attr:{self._table.physical_name()}_{self.physical_name()}')

    def ref(self):
        return ColumnRef(self._table.name, self._metadata.name)


class SnowflakeTable(BindableTable):
    _name: str
    _columns: dict[str, BindableColumn]
    _foreign_keys: set[ForeignKey]

    def __init__(self, metadata: 'TabularMetadata', model):
        self._name = metadata.name
        self._columns = {col.name: BindableColumn(col, self, model) for col in metadata.columns}
        self._foreign_keys = metadata.foreign_keys
        for fk in self._foreign_keys:
            # TODO : this doesn't work for composite FKs
            for col in fk.source_columns:
                target_col = fk.target_columns[0]
                self._columns[col.column].references = target_col

    def __getattr__(self, key):
        if key in self.__dict__:
            return self.__dict__[key]
        if key in self._columns:
            return self._columns[key]
        else:
             raise AttributeError(f'Snowflake table "{self._name}" has no column named "{key}"')

    def __str__(self):
        # returns the name of the table, as well as the columns and their types
        return self._name + ':\n' + '\n'.join(
            [f' {col.metadata.name} {col.metadata.datatype}' for _, col in self._columns.items()]
        ) + '\n' + '\n'.join(
            [f' {fk.source_columns} -> {fk.target_columns}' for fk in self._foreign_keys]
        )

    @property
    def name(self):
        return self._name

    def physical_name(self):
        # physical relation name is always in the form of `{database}_{schema}_{table}
        return self._name.lower().replace('.', '_')

    def key_type(self):
        return RowId



import json
import re
from collections import defaultdict

import xmltodict

from relationalai.early_access.dsl.adapters.orm.model import ORMEntityType, ORMValueType, ORMRole, SubtypeArrow, \
    ORMUniquenessConstraint


class ORMParser:

    def __init__(self, orm_file_path):
        with open(orm_file_path) as orm_file:
            data = xmltodict.parse(orm_file.read())

        self._ontology = json.loads(json.dumps(data))

        self._model_name = self._parse_model_name()
        self._entity_types = self._parse_entity_types()
        self._value_types = self._parse_value_types()
        self._object_types = {**self._value_types, **self._entity_types}
        self._roles = self._parse_roles()
        self._role_value_constraints = self._parse_role_value_constraints()
        self._subtype_facts = self._parse_subtype_facts()
        self._internal_uniqueness_constraints, self._external_uniqueness_constraints = self._parse_uniqueness_constraints()
        self._fact_type_readings = self._parse_fact_types_reading_orders()

    def model_name(self):
        return self._model_name

    def entity_types(self) -> dict[str, ORMEntityType]:
        return self._entity_types

    def object_types(self):
        return self._object_types

    def value_types(self):
        return self._value_types

    def roles(self):
        return self._roles

    def role_value_constraints(self):
        return self._role_value_constraints

    def subtype_facts(self):
        return self._subtype_facts

    def internal_uniqueness_constraints(self):
        return self._internal_uniqueness_constraints

    def external_uniqueness_constraints(self):
        return self._external_uniqueness_constraints

    def fact_type_readings(self):
        return self._fact_type_readings

    def fact_type_to_roles(self):
        fact_type_data = defaultdict(list)
        for role in self._roles.values():
            relationship_name = role.relationship_name
            fact_type_data[relationship_name].append(role)
        return fact_type_data

    def fact_type_to_unique_constraints(self):
        fact_type_to_uc = defaultdict(list)

        for uc_id, uc in self._internal_uniqueness_constraints.items():
            role_id = uc.roles[0]
            role = self._roles.get(role_id)

            if not role or not role.relationship_name:
                continue

            fact_type_to_uc[role.relationship_name].append(uc_id)

        return fact_type_to_uc

    def _parse_model_name(self):
        model_name = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "@Name")
        return model_name if model_name else "ORMModel"

    def _parse_entity_types(self):
        entity_types = {}
        orm_entity_types = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                            "orm:Objects", "orm:EntityType")
        if orm_entity_types:
            for et in orm_entity_types:
                id = et["@id"]
                name = et['@Name']
                reference_mode = et.get("@_ReferenceMode")
                preferred_id = None if reference_mode == "" else name + "_" + reference_mode
                entity_types[id] = ORMEntityType(id, name, preferred_id)
        return entity_types

    def _parse_value_types(self):
        value_types = {}
        data_types = self._parse_data_types()
        orm_value_types = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                            "orm:Objects", "orm:ValueType")
        if orm_value_types:
            for vt in orm_value_types:
                id = vt["@id"]
                name = vt['@Name']
                data_type = data_types[vt["orm:ConceptualDataType"]["@ref"]]
                value_types[id] = ORMValueType(id, name, data_type)
        return value_types

    def _parse_data_types(self):
        data_types = {}
        orm_data_types = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "orm:DataTypes")
        if orm_data_types:
            for k, v in orm_data_types.items():
                if v is not None:
                    data_types[v["@id"]] = k[4:]
        return data_types

    def _parse_roles(self):
        roles = {}
        role_to_player = self._parse_role_player()
        orm_fact_types = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "orm:Facts", "orm:Fact")
        if orm_fact_types:
            for ft in orm_fact_types:
                relationship_name = (ft['@_Name'])
                orm_roles = self._get_nested(ft, "orm:FactRoles", "orm:Role")
                if orm_roles:
                    roles_list = orm_roles if isinstance(orm_roles, list) else [orm_roles]
                    for ro in roles_list:
                        role_id = ro["@id"]
                        role_name = ro["@Name"]
                        roles[role_id] = ORMRole(role_id, role_name, relationship_name, role_to_player[role_id])
        return roles

    def _parse_role_player(self):
        role_to_player = {}
        object_types = ((self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "orm:Objects", "orm:EntityType") or []) +
                        (self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "orm:Objects", "orm:ValueType") or []))
        for ot in object_types:
            roles = self._get_nested(ot, "orm:PlayedRoles", "orm:Role")
            if roles:
                roles_list = roles if isinstance(roles, list) else [roles]
                for role in roles_list:
                    role_to_player[role["@ref"]] = ot["@id"]
        return role_to_player

    def _parse_role_value_constraints(self):
        role_value_constraints = defaultdict(list)
        fact_types = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "orm:Facts", "orm:Fact")
        if fact_types:
            for ft in fact_types:
                roles = self._get_nested(ft,"orm:FactRoles", "orm:Role")
                if roles and isinstance(roles, list):
                    for ro in roles:
                        value_ranges = self._get_nested(ro, "orm:ValueRestriction", "orm:RoleValueConstraint",
                                                        "orm:ValueRanges", "orm:ValueRange")
                        if value_ranges:
                            for rvc in value_ranges:
                                if rvc.get("@MinValue") != rvc.get("@MaxValue"):
                                    raise Exception("Unsupported value range constraint.")
                                role_value_constraints[ro["@id"]].append(rvc["@MinValue"])
        return role_value_constraints

    def _parse_subtype_facts(self):
        subtype_of = {}
        object_types = self._object_types
        for subtype_arrow in self._parse_subtype_arrows():
            subtype = object_types[subtype_arrow.start]
            supertype = object_types[subtype_arrow.end]
            if subtype_of.get(supertype.name) is not None:
                subtype_of[supertype.name].append(subtype.name)
            else:
                subtype_of[supertype.name] = [subtype.name]
        return subtype_of

    def _parse_subtype_arrows(self):
        subtype_of = []
        orm_subtype_facts = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                             "orm:Facts", "orm:SubtypeFact")
        if orm_subtype_facts:
            subtype_facts = orm_subtype_facts if isinstance(orm_subtype_facts, list) else [orm_subtype_facts]
            for sft in subtype_facts:
                subtype_of.append(self._parse_subtype_arrow(sft))
        return subtype_of

    def _parse_subtype_arrow(self, subtype_fact):
        fact_roles = subtype_fact["orm:FactRoles"]
        subtype = self._get_nested(fact_roles, "orm:SubtypeMetaRole", "orm:RolePlayer", "@ref")
        supertype = self._get_nested(fact_roles, "orm:SupertypeMetaRole", "orm:RolePlayer", "@ref")
        return SubtypeArrow(subtype, supertype)

    def _parse_uniqueness_constraints(self):
        internal = {}
        external = {}
        orm_ucs = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel",
                                   "orm:Constraints", "orm:UniquenessConstraint")
        if orm_ucs:
            for uc in orm_ucs:
                uc_id = uc["@id"]
                pid = uc.get("orm:PreferredIdentifierFor", None)
                identifies = None
                if pid is not None and pid["@ref"] in self._entity_types:
                    identifies = pid["@ref"]
                uc_roles = []
                roles = self._get_nested(uc, "orm:RoleSequence", "orm:Role")
                if roles:
                    roles_list = roles if isinstance(roles, list) else [roles]
                    for ro in roles_list:
                        role_id = ro["@ref"]
                        uc_roles.append(role_id)
                constraint = ORMUniquenessConstraint(uc_id, uc_roles, identifies)
                target = internal if uc.get("@IsInternal") == "true" else external
                target[uc_id] = constraint
        return internal, external

    def _parse_fact_types_reading_orders(self):
        fact_types_readings = defaultdict(list)
        fact_types = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "orm:Facts", "orm:Fact")
        if fact_types:
            for ft in fact_types:
                relationship_name = (ft['@_Name'])
                if ft.get("orm:ReadingOrders"):
                    reading_orders = self._get_nested(ft, "orm:ReadingOrders", "orm:ReadingOrder")
                    if reading_orders:
                        ros = reading_orders if isinstance(reading_orders, list) else [reading_orders]
                        for ro in ros:
                            reading = self._parse_fact_types_reading_order(ro)
                            if reading:
                                fact_types_readings[relationship_name].append(self._parse_reading(reading))
        return fact_types_readings

    def _parse_fact_types_reading_order(self, reading_order):
        reading = self._get_nested(reading_order, "orm:Readings", "orm:Reading", "orm:Data")
        if reading:
            roles = self._get_nested(reading_order, "orm:RoleSequence", "orm:Role")
            if roles:
                roles_list = roles if isinstance(roles, list) else [roles]
                for i, seq in enumerate(roles_list):
                    role_id = seq["@ref"]
                    player = self._roles[role_id].player
                    reading = reading.replace(f"{{{i}}}", f"{{{player}}}")
            return reading
        return None

    @staticmethod
    def _parse_reading(reading):
        pattern = r'(\{_[0-9A-Fa-f\-]{36}\})\s*(.*?)\s*(?=\{_[0-9A-Fa-f\-]{36}\}|$)'
        matches = re.findall(pattern, reading)
        players = []
        texts = []

        for match in matches:
            player, text = match
            players.append(player.strip('{}'))
            texts.append(text.strip())
        return {'players': players, 'texts': texts}

    @staticmethod
    def _get_nested(d, *keys):
        for key in keys:
            d = d.get(key)
            if d is None:
                return None
        return d


from .core import std
from .core.exprs import _
from .core.relations import rule, addrule, AbstractRelation, Relation, ExternalRelation, EntityPopulationRelation, \
    RelationSignature
from .core.namespaces import Namespace
from .core.rules import Vars, Annotation
from .core.types.standard import BigInteger, PositiveInteger, Hash, Boolean, Decimal, UnsignedDecimal, PositiveDecimal, \
    Float, BigUnsignedInteger, BigPositiveInteger, Integer, UnsignedInteger, Date, DateTime, RowId, String, Any, Symbol
from .core.logic.aggregation import Aggregation
from .ir.executor import RelExecutor
from .ir.compiler import Compiler
from .ontologies.relationships import Relationship
from .ontologies.subtyping import SubtypeArrow
from .ontologies.models import Model
from .types.entities import EntityType
from .types.values import ValueType, ValueSubtype
from .relations import Seq, Seq0, Iseq, Iseq0
from .schemas import ConjunctiveSchema, DisjunctiveSchema, FalseSchema
from .physical_metadata.tabular_source_module import TabularSourceModule
from .physical_metadata.tabular_source import TabularSource

__all__ = ['std', '_', 'rule', 'addrule', 'Vars', 'Annotation', 'BigInteger', 'PositiveInteger', 'Hash', 'Boolean',
           'Decimal', 'UnsignedDecimal', 'PositiveDecimal', 'Float', 'BigUnsignedInteger', 'BigPositiveInteger',
           'Integer', 'UnsignedInteger', 'Date', 'DateTime', 'RowId', 'String', 'Any', 'Symbol', 'Model',
           'TabularSource', 'Aggregation', 'Namespace', 'TabularSourceModule', 'ConjunctiveSchema', 'DisjunctiveSchema',
           'FalseSchema', 'Seq', 'Seq0', 'Iseq', 'Iseq0', 'RelExecutor', 'Compiler', 'AbstractRelation', 'Relation',
           'ExternalRelation','EntityPopulationRelation', 'RelationSignature', 'Relationship', 'SubtypeArrow',
           'EntityType', 'ValueType', 'ValueSubtype']



"""
A simple metamodel for Rel.
"""
from __future__ import annotations
from dataclasses import dataclass, field
from io import StringIO
from typing import Optional, Tuple, Union as PyUnion
from relationalai.early_access.rel import rel_utils
from relationalai.early_access.metamodel.util import Printer as BasePrinter

@dataclass(frozen=True)
class Node:
    def __str__(self):
        return to_string(self)

    @property
    def kind(self):
        return self.__class__.__name__.lower()

#--------------------------------------------------
# Top level program and declarations
#--------------------------------------------------

@dataclass(frozen=True)
class Program(Node):
    # top-level declarations
    declarations: Tuple[PyUnion[Declare, Def], ...]

@dataclass(frozen=True)
class Declare(Node):
    """ declare $premise [requires $requires] """
    premise: Expr
    requires: Optional[Expr]
    annotations: Tuple[Annotation, ...] = field(default_factory=tuple)

@dataclass(frozen=True)
class Def(Node):
    """ def $name($params) { $body } """
    name: str
    params: Tuple[PyUnion[Var, Primitive, MetaValue], ...]
    body: Expr
    annotations: Tuple[Annotation, ...] = field(default_factory=tuple)

#--------------------------------------------------
# Primitives, Annotations
#--------------------------------------------------

Primitive = PyUnion[str, int, float, bool]

@dataclass(frozen=True)
class Annotation(Node):
    """ @$name($args) """
    name: str
    args: Tuple[PyUnion[Primitive, MetaValue], ...]


#--------------------------------------------------
# Expr
#--------------------------------------------------

@dataclass(frozen=True)
class CompositeExpr(Node):
    pass

Expr = PyUnion[Primitive, CompositeExpr]

@dataclass(frozen=True)
class MetaValue(CompositeExpr):
    """ #$value """
    value: Primitive

@dataclass(frozen=True)
class Var(CompositeExpr):
    """ $name[...] [in $type] """
    name: str
    varargs: bool = False
    type: Optional[str] = None

@dataclass(frozen=True)
class Identifier(CompositeExpr):
    """ $name

    Used to declare a relation by name or to refer to a relation (e.g. Int, String)
    """
    name: str

@dataclass(frozen=True)
class Atom(CompositeExpr):
    """ $expr($args)

    Represents atoms like identifier($name)($args) as well as literal relations
    relations like {(:a, Int, 1)}(x...)
    """
    expr: Expr
    args: Tuple[Expr, ...]

def atom(name: str, args: Tuple[Expr, ...]):
    """ Helper to create a Atom where expression is an identifier. """
    return Atom(Identifier(name), args)

@dataclass(frozen=True)
class RelationalAbstraction(CompositeExpr):
    """ ($head): $body """
    head: Tuple[Expr, ...]
    body: Expr

@dataclass(frozen=True)
class And(CompositeExpr):
    """ $body[0] and $body[1] ... and $body[n] """
    body: Tuple[Expr, ...]

@dataclass(frozen=True)
class Or(CompositeExpr):
    """ $body[0] or $body[1] ... or $body[n] """
    body: Tuple[Expr, ...]

@dataclass(frozen=True)
class Exists(CompositeExpr):
    """ exists(($vars) | $body ) """
    vars: Tuple[Var, ...]
    body: Expr

@dataclass(frozen=True)
class ForAll(CompositeExpr):
    """ forall(($vars) | $body ) """
    vars: Tuple[Var, ...]
    body: Expr

@dataclass(frozen=True)
class Not(CompositeExpr):
    """ not ( $body ) """
    body: Expr

@dataclass(frozen=True)
class BinaryExpr(CompositeExpr):
    """ $lhs $op $rhs """
    lhs: Expr
    op: str
    rhs: Expr

@dataclass(frozen=True)
class Product(CompositeExpr):
    """ ($body[0] , $body[1] ... , $body[n]) """
    body: Tuple[Expr, ...]

@dataclass(frozen=True)
class Union(CompositeExpr):
    """ {$body[0] ; $body[1] ... ; $body[n]} """
    body: Tuple[Expr, ...]



#--------------------------------------------------
# Printer
#--------------------------------------------------

infix = ["+", "-", "*", "/", "%", "=", "!=", "<", "<=", ">", ">="]

def to_string(node) -> str:
    io = StringIO()
    Printer(io).print_node(node, 0)
    return io.getvalue()

@dataclass(frozen=True)
class Printer(BasePrinter):

    def _join(self, args, sep=', ', indent=0):
        for i, s in enumerate(args):
            if i != 0:
                if indent != 0:
                    self._print(sep.rstrip())
                    self._nl()
                else:
                    self._print(sep)
            self.print_node(s, indent)

    def _print_value(self, value):
        if isinstance(value, tuple):
            for i, v in enumerate(value):
                if i != 0:
                    self._print(", ")
                self._print_value(v)
        elif isinstance(value, str):
            self._print(f"\"{value}\"")
        elif isinstance(value, bool):
            self._print(str(value).lower())
        else:
            self._print(str(value))

    def print_node(self, node, indent=0) -> None:
        #--------------------------------------------------
        # Top level program and declarations
        #--------------------------------------------------
        if isinstance(node, Program):
            for d in node.declarations:
                self.print_node(d, indent)
                self._nl()

        elif isinstance(node, Declare):
            for anno in node.annotations:
                self.print_node(anno, indent)
                self._nl()
            self._indent_print(indent, f"declare {node.premise}")
            if node.requires:
                self._print(" requires ")
                self.print_node(node.requires, 0)
            self._nl()

        elif isinstance(node, Def):
            for anno in node.annotations:
                self.print_node(anno, indent)
                self._nl()
            self._indent_print(indent, f"def {node.name}")
            if node.params:
                self._print("(")
                self._join(node.params)
                self._print("):")
                self._nl()
            self.print_node(node.body, indent+1)
            self._nl()

        #--------------------------------------------------
        # Primitives, Annotations
        #--------------------------------------------------
        elif isinstance(node, (str, int, float, bool, tuple)):
            self._print_value(node)

        elif isinstance(node, Annotation):
            self._indent_print(indent, f"@{node.name}")
            if node.args:
                self._print("(")
                self._join(node.args)
                self._print(")")


        #--------------------------------------------------
        # Expr
        #--------------------------------------------------

        elif isinstance(node, MetaValue):
            if isinstance(node.value, str):
                self._print(":")
            else:
                self._print("#")
            self._print_value(node.value)

        elif isinstance(node, Var):
            self._indent_print(indent, f"{node.name}")
            if node.varargs:
                self._print("...")
            if node.type and node.type != "Any":
                self._print(f" in {node.type}")

        elif isinstance(node, Identifier):
            self._indent_print(indent, f"{node.name}")

        elif isinstance(node, Atom):
            if isinstance(node.expr, Identifier) and node.expr.name in infix:
                # deal with the 3 kinds of infix operators
                if len(node.args) == 1:
                    self.print_node(node.expr, indent)
                    self.print_node(node.args[0])
                elif len(node.args) == 2:
                    self.print_node(node.args[0], indent)
                    self._print(" ")
                    self.print_node(node.expr)
                    self._print(" ")
                    self.print_node(node.args[1])
                elif len(node.args) == 3:
                    self.print_node(node.args[2], indent)
                    self._print(" = ")
                    self.print_node(node.args[0])
                    self._print(" ")
                    self.print_node(node.expr)
                    self._print(" ")
                    self.print_node(node.args[1])
                else:
                    raise NotImplementedError(f"emit_action: {node}")
            else:
                is_agg = False
                if isinstance(node.expr, Identifier):
                    self.print_node(node.expr, indent)
                    is_agg = node.expr.name in rel_utils.OPERATORS.values()
                else:
                    self._indent_print(indent, "{")
                    self.print_node(node.expr, indent)
                    self._print("}")
                self._print("(")
                if is_agg:
                    self._nl()
                    self._join(node.args, indent=indent+1)
                    self._nl()
                    self._indent_print(indent, ")")
                else:
                    self._join(node.args)
                    self._print(")")

        elif isinstance(node, RelationalAbstraction):
            if node.head:
                self._indent_print(indent, "{(")
                self._join(node.head)
                self._print("):")
                self._nl()
                self.print_node(node.body, indent+1)
                self._nl()
                self._indent_print(indent, "}")
            else:
                self.print_node(node.body, indent)

        elif isinstance(node, And):
            self._join(node.body, " and ", indent)

        elif isinstance(node, Or):
            self._join(node.body, " or ", indent)

        elif isinstance(node, Exists):
            self._indent_print(indent, "exists((")
            self._join(node.vars)
            self._print(") |")
            self._nl()
            self.print_node(node.body, indent+1)
            self._nl()
            self._indent_print(indent, ")")

        elif isinstance(node, ForAll):
            self._indent_print(indent, "forall((")
            self._join(node.vars)
            self._print(")| ")
            self.print_node(node.body)
            self._print(")")

        elif isinstance(node, Not):
            self._indent_print(indent, "not (")
            self.print_node(node.body)
            self._print(")")

        elif isinstance(node, BinaryExpr):
            self.print_node(node.lhs, indent)
            self._print(f" {node.op} ")
            self.print_node(node.rhs)

        elif isinstance(node, Product):
            self._indent_print(indent, "(")
            self._join(node.body, ", ")
            self._print(")")

        elif isinstance(node, Union):
            self._indent_print(indent, "{")
            self._join(node.body, " ; ")
            self._print("}")

        else:
            raise Exception(f"Missing implementation in Rel printer: {type(node)}")



"""
    Functions to simplify the creation of IR nodes using some common variations.
"""
from __future__ import annotations
from typing import Any, Tuple, Optional, Sequence as PySequence, Union
import decimal

from relationalai.early_access.metamodel import ir, types
from relationalai.early_access.metamodel.util import OrderedSet, FrozenOrderedSet, frozen, ordered_set
from relationalai.early_access.metamodel.visitor import collect_by_type


#-------------------------------------------------
# Public Types - Model
#-------------------------------------------------

def model(
        engines: OrderedSet[ir.Engine],
        relations: OrderedSet[ir.Relation],
        types: OrderedSet[ir.Type],
        root: ir.Task):
    return ir.Model(
        engines.frozen(),
        relations.frozen(),
        types.frozen(),
        root
    )

def compute_model(root: ir.Task) -> ir.Model:
    return model(
        collect_by_type(ir.Engine, root),
        collect_by_type(ir.Relation, root),
        _collect_reachable_types(root),
        root
    )

def _collect_reachable_types(root: ir.Task) -> OrderedSet[ir.Type]:
    """ Collect all types reachable from this task, including super types, element types, etc. """
    # add all type explicitly used in the root task
    start = collect_by_type(ir.Type, root)

    # add all built-in types
    start.update(types.builtin_types)

    # add super types and element types of collections
    to_visit = start.list.copy()
    while to_visit:
        t = to_visit.pop()
        start.add(t)
        if isinstance(t, ir.ScalarType):
            to_visit.extend(t.super_types)
        elif isinstance(t, (ir.ListType, ir.SetType)):
            to_visit.append(t.element_type)
        else:
            # UnionType
            to_visit.extend(t.types)
    return start


def install(model: ir.Model, logical_task: ir.Logical) -> ir.Model:
    """ Return a new model that represents this model with the logical task as the root task. """

    # the model's root is interpreted as the "installed model", we want a model that
    # computes those tasks and succeeds only if the incoming logical task l succeeds. To do
    # so, we create a new model whose root task is a logical; this logical succeeds if l
    # succeeds, taking into account the installed model. So we wrap the installed model in a
    # union with an empty Task, which always succeeds, to guarantee that the installed model
    # always "succeeds".

    # TODO: when installing a logical, we may need to bring new engines, relations and types;
    # we can either extend this function's interface to receive that (basically a new model
    # to be merged), or we can traverse `l` to lookup engines/relations/types to add to the
    # new model.
    # TODO: is it OK to make an empty Task represent "success"?
    return ir.Model(
        model.engines, model.relations, model.types,
        logical(
            [logical_task, union([model.root, ir.Task(None)])]
        )
    )

def install_logical(rules: ir.Logical, select: ir.Logical, engine: Optional[ir.Engine]=None) -> ir.Logical:
    """ Similar to install but only manipulating the tasks."""
    # TODO - this may not be necessary because it is OK if the `select` logical fails, it just
    # returns false, i.e. no values.
    return logical([
        select,
        union([rules, ir.Task(None)])
    ], [], engine)

#-------------------------------------------------
# Public Types - Engine
#-------------------------------------------------

def capability(name: str):
    return ir.Capability(name)

def engine(name: str, platform: str, relations:OrderedSet[ir.Relation], info: Any=None, capabilities: OrderedSet[ir.Capability]=ordered_set()):
    return ir.Engine(name, platform, info, capabilities.frozen(), relations.frozen())

#-------------------------------------------------
# Public Types - Data
#-------------------------------------------------

def scalar_type(name: str, super_types:list[ir.ScalarType]=[]):
    return ir.ScalarType(name, FrozenOrderedSet(super_types))

def list_type(element_type: ir.Type):
    return ir.ListType(element_type)

def set_type(element_type: ir.Type):
    return ir.SetType(element_type)

def union_type(types: list[ir.Type]):
    return ir.UnionType(FrozenOrderedSet(types))

def field(name: str, type: ir.Type, input: bool=False):
    return ir.Field(name, type, input)

def input_field(name:str, type: ir.Type):
    return ir.Field(name, type, True)


# property helpers
def relation(name: str, fields: list[ir.Field], requires: OrderedSet[ir.Capability]=ordered_set(), annos: Optional[PySequence[ir.Annotation]]=None):
    return ir.Relation(name, tuple(fields), requires.frozen(), frozen() if annos is None else FrozenOrderedSet(annos))

def entity(type: ir.ScalarType):
    return relation(type.name, [field("id", type)])

def property(name: str, key_name: str, key_type: ir.Type, value_name:str, value_type: ir.Type):
    return relation(
        name,
        [field(key_name, key_type), field(value_name, value_type)]
    )

#-------------------------------------------------
# Public Types - Tasks
#-------------------------------------------------

def success():
    return ir.Task(None)

#
# Task composition
#

def logical(body: PySequence[ir.Task], hoisted: PySequence[ir.VarOrDefault]=[], engine: Optional[ir.Engine]=None):
    return ir.Logical(engine, tuple(hoisted), tuple(body))

def sequence(tasks: PySequence[ir.Task], hoisted: PySequence[ir.VarOrDefault]=[], engine: Optional[ir.Engine]=None):
    return ir.Sequence(engine, tuple(hoisted), tuple(tasks))

def union(tasks: PySequence[ir.Task], hoisted: PySequence[ir.VarOrDefault]=[], engine: Optional[ir.Engine]=None):
    return ir.Union(engine, tuple(hoisted), tuple(tasks))

def match(tasks: PySequence[ir.Task], hoisted: PySequence[ir.VarOrDefault]=[], engine: Optional[ir.Engine]=None):
    return ir.Match(engine, tuple(hoisted), tuple(tasks))

def until(check: ir.Task, body: ir.Task, hoisted: PySequence[ir.VarOrDefault]=[], engine: Optional[ir.Engine]=None):
    return ir.Until(engine, tuple(hoisted), check, body)

def wait(check: ir.Task, hoisted: PySequence[ir.VarOrDefault]=[], engine: Optional[ir.Engine]=None):
    return ir.Wait(engine, tuple(hoisted), check)

#
# Relational Operations
#

def var(name: str, type: Union[ir.Type, None] = None) -> ir.Var:
    if type is None:
        return ir.Var(types.Any, name)
    else:
        return ir.Var(type, name)

def default(var: ir.Var, value: ir.Value) -> ir.Default:
    return ir.Default(var, value)

def wild() -> ir.Var:
    return var("_", types.Any)

def literal(value: Any, type: Union[ir.Type, None] = None) -> ir.Value:
    """ Create a Literal with this type. If type is not present, attempt to figure out the
    appropriate type. See: `l`. """
    if type is None:
        return lit(value)
    else:
        return ir.Literal(type, value)

def lit(value: Any) -> ir.Value:
    """ Ensure this value is an appropriate ir.Value. This function wraps common python values
     in ir.Literals, and supports lists of values. """
    if isinstance(value, ir.Literal):
        return value
    elif isinstance(value, str):
        return ir.Literal(types.String, value)
    elif isinstance(value, int):
        return ir.Literal(types.Int, value)
    elif isinstance(value, float):
        return ir.Literal(types.Decimal, value)
    elif isinstance(value, bool):
        return ir.Literal(types.Bool, value)
    elif isinstance(value, decimal.Decimal):
        return ir.Literal(types.Decimal, float(value))
    elif isinstance(value, list):
        return tuple([lit(v) for v in value])
    else:
        raise NotImplementedError(f"literal value: {value} of type {type(value)}")

def annotation(relation: ir.Relation, args: PySequence[ir.Value]) -> ir.Annotation:
    return ir.Annotation(relation, tuple(args))

def derive(relation: ir.Relation, args: PySequence[ir.Value], annos: Optional[PySequence[ir.Annotation]]=None, engine: Optional[ir.Engine]=None):
    return ir.Update(engine, relation, tuple(args), ir.Effect.derive, frozen() if annos is None else FrozenOrderedSet(annos))

def insert(relation: ir.Relation, args: PySequence[ir.Value], annos: Optional[PySequence[ir.Annotation]]=None, engine: Optional[ir.Engine]=None):
    return ir.Update(engine, relation, tuple(args), ir.Effect.insert, frozen() if annos is None else FrozenOrderedSet(annos))

def delete(relation: ir.Relation, args: PySequence[ir.Value], annos: Optional[PySequence[ir.Annotation]]=None, engine: Optional[ir.Engine]=None):
    return ir.Update(engine, relation, tuple(args), ir.Effect.delete, frozen() if annos is None else FrozenOrderedSet(annos))

def lookup(relation: ir.Relation, args: PySequence[ir.Value], engine: Optional[ir.Engine]=None):
    return ir.Lookup(engine, relation, tuple(args))

def update(relation: ir.Relation, args: PySequence[ir.Value], effect: ir.Effect, annos: Optional[PySequence[ir.Annotation]]=None, engine: Optional[ir.Engine]=None):
    return ir.Update(engine, relation, tuple(args), effect, frozen() if annos is None else FrozenOrderedSet(annos))

def output(vars: PySequence[Union[ir.Var, Tuple[str, ir.Var]]], engine: Optional[ir.Engine]=None):
    """Create an output task that will return values bound to these variables. The vars
    sequence can contain plain Vars, in which case the alias used in the output is the
    variable name. This alias can be customized by adding a tuple (alias, Var) to vars.
    """
    s = ordered_set()
    for x in vars:
        if isinstance(x, ir.Var):
            s.add((x.name, x))
        else:
            s.add(x)
    return ir.Output(engine, s.frozen())


def construct(var: ir.Var, bindings: dict[ir.Relation, Any], engine: Optional[ir.Engine]=None):
    """Create a Construct node that will create an id and bind to this var. The id will be a
    hash of the types of the var, followed by the name and value of the bindings."""
    values = []
    values.append(var.type)
    for relation, value in bindings.items():
        values.append(ir.Literal(types.String, relation.name))
        values.append(value)
    return ir.Construct(engine, tuple(values), var)

def aggregate(aggregation: ir.Relation, projection: PySequence[ir.Var], group: PySequence[ir.Var], args: PySequence[Any], engine: Optional[ir.Engine]=None):
    return ir.Aggregate(engine, aggregation, tuple(projection), tuple(group), tuple(args))


#
# Logical Quantifiers
#

def not_(task: ir.Task, engine: Optional[ir.Engine]=None):
    return Not(task, engine)

def Not(task: ir.Task, engine: Optional[ir.Engine]=None):
    return ir.Not(engine, task)

def exists(vars: PySequence[ir.Var], task: ir.Task, engine: Optional[ir.Engine]=None):
    return ir.Exists(engine, tuple(vars), task)

def for_all(vars: PySequence[ir.Var], task: ir.Task, engine: Optional[ir.Engine]=None):
    return ir.ForAll(engine, tuple(vars), task)


#
# Iteration (Loops)
#

# loops body until a break condition is met
def loop(iter: ir.Var, body: ir.Task, hoisted: PySequence[ir.VarOrDefault]=[], engine: Optional[ir.Engine]=None):
    return ir.Loop(engine, tuple(hoisted), iter, body)

def break_(check: ir.Task, engine: Optional[ir.Engine]=None):
    return Break(check, engine)

def Break(check: ir.Task, engine: Optional[ir.Engine]=None):
    return ir.Break(engine, check)



"""
Builder API for RelationalAI.
"""

from relationalai.early_access.builder.builder import (
    Model, Concept, Relationship, Expression, Fragment, Error,
    String, Integer,
    select, where, require, then, distinct, union,
    count, sum, min, max, avg, per,
    not_, forall, exists
)

__all__ = [
    "Model", "Concept", "Relationship", "Expression", "Fragment", "Error",
    "String", "Integer",
    "select", "where", "require", "then", "distinct", "union",
    "count", "sum", "min", "max", "avg", "per",
    "not_", "forall", "exists"
]



from copy import deepcopy
from logging import FileHandler
import os
import sys
from uuid import UUID
from typing import Dict, Union

from relationalai.debugging import JsonFormatter
from relationalai.util import get_timestamp
from relationalai.util.format import str_to_bool
from relationalai.util.span_tracker import SpanTracker, record_is_span

# Span types that are automatically generated by PyRel but are not relevant for performance analysis
SKIP_SPAN_TYPES = ['check_status']


class SpansFileHandler(FileHandler):
    """
    This logger writes spans to a local .jsonl file inside the `spans_output` folder as they finish.
    The primary goal is to ingest these spans into Snowflake for analytics.
    """

    def __init__(self, trace_id: UUID, filename: Union[str,None] = None):
        self.span_tracker = SpanTracker(trace_id, span_types_to_skip=SKIP_SPAN_TYPES, log_span_attrs_as_str=False)
        if filename is None:
            filename = self._get_default_log_file_path()
            if not os.path.exists(os.path.dirname(filename)):
                os.makedirs(os.path.dirname(filename))
        FileHandler.__init__(self, filename=filename)
        self.setFormatter(JsonFormatter())

    def emit(self, record):
        if record_is_span(record):
            msg: Dict = record.msg # type: ignore
            if msg["event"] == "span_start":
                self.span_tracker.handle_span_start(msg)
            if msg["event"] == "span_end":
                span_end_result = self.span_tracker.handle_span_end(msg)
                if span_end_result:
                    formatted_span_msg = span_end_result[0]
                    if formatted_span_msg is None: # This will happen for span_types_to_skip
                        return
                    record_copy = deepcopy(record)  # Create a deep copy so as to not interfere with other handlers
                    record_copy.msg = formatted_span_msg
                    FileHandler.emit(self, record_copy)

    def _get_default_log_file_path(self) -> str:
        now_str = get_timestamp().strftime("%Y%m%d_%H%M%S")
        trace_id_trunc = str(self.span_tracker.trace_id)[:8]
        spans_subfolder = os.getenv("SPANS_SUBFOLDER", "")
        return os.path.join(
            os.getcwd(), "output_spans", spans_subfolder, f"{now_str}__{trace_id_trunc}.jsonl"
        )


def is_collect_only():
    return "pytest" in sys.modules and "--collect-only" in sys.argv


def maybe_add_spans_file_handler(logger, trace_id:UUID, if_not_set: bool = False):
    """
    Add a SpansFileHandler to the logger if the 'ENABLE_SPANS_FILE_HANDLER' environment variable is set to True,
    or if the 'if_not_set' parameter is True.
    """
    if str_to_bool(os.getenv("ENABLE_SPANS_FILE_HANDLER")) or if_not_set:
        if is_collect_only():
            print("Running in collect-only mode, skipping SpansFileHandler")
        elif any(isinstance(handler, SpansFileHandler) for handler in logger.handlers):
            print("logger already has a SpansFileHandler instance, won't add another one")
        else:
            logger.debug("Adding SpansFileHandler to logger")
            logger.addHandler(SpansFileHandler(trace_id))



from __future__ import annotations
from dataclasses import dataclass, field
from typing import Optional, Tuple, cast

from relationalai.early_access.metamodel import ir, compiler as c, visitor as v, factory as f, builtins, types
from relationalai.early_access.metamodel.util import FrozenOrderedSet, OrderedSet, frozen, ordered_set, filter_by_type
from relationalai.early_access.devmode import sql, rewrite


class Compiler(c.Compiler):
    def __init__(self):
        super().__init__([
            rewrite.Denormalize()
            # group updates
            # compute SCCs, use Sequence to denote their order
        ])

    def do_compile(self, model: ir.Model, options:dict={}) -> str:
        # print()
        # print(model)
        v = ModelToSQL()
        model.accept(v)
        # print()
        # print(v.to_sql_code())
        return v.to_sql_code()


@dataclass
class ModelToSQL(v.Visitor):
    """ Generates SQL from an IR Model, assuming the compiler rewrites were done. """

    # the resulting statements
    statements: list[sql.Node] = field(default_factory=list)

    #--------------------------------------------------
    # Public API
    #--------------------------------------------------

    def to_sql_code(self):
        return "\n".join(str(s) for s in self.statements)


    #--------------------------------------------------
    # Visitor Implementation
    #--------------------------------------------------

    def visit(self, n: ir.Node, parent: Optional[ir.Node]):
        # declare tables when visiting the relation list in the model
        if isinstance(n, ir.Relation) and isinstance(parent, ir.Model) and not builtins.is_builtin(n):
            self._create_table(n)
            return False

        if isinstance(n, ir.Union):
            # currently treat as a recursive view
            self._create_recursive_view(n)
            return False

        if isinstance(n, ir.Logical) and isinstance(parent, ir.Logical):
            # currently only take the logical tasks under the top-level logical
            self._create_statement(n)
            return False
        return True

    #--------------------------------------------------
    # SQL Generation
    #--------------------------------------------------
    def _create_table(self, r: ir.Relation):
        self.statements.append(sql.CreateTable(
            sql.Table(r.name,
                list(map(lambda f: sql.Column(f.name, self._convert_type(f.type)), r.fields))
            )))

    def _create_recursive_view(self, union: ir.Union):
        assert len(union.tasks) == 2
        assert isinstance(union.tasks[0], ir.Logical)
        assert isinstance(union.tasks[1], ir.Logical)

        def make_case_select(logical: ir.Logical):
            # TODO - improve the typing info to avoid these casts
            lookups = cast(list[ir.Lookup], filter_by_type(logical.body, ir.Lookup))
            # TODO - assuming a single update per case
            update = v.collect_by_type(ir.Update, logical).some()

            # TODO - rewriting references to the view, to use the CTE instead, with _rec
            new_lookups = []
            for lookup in lookups:
                if lookup.relation == update.relation:
                    new_lookups.append(f.lookup(
                        ir.Relation(f"{lookup.relation.name}_rec", lookup.relation.fields, f.frozen(), f.frozen()), lookup.args, lookup.engine))
                else:
                    new_lookups.append(lookup)

            aliases = []
            for i, arg in enumerate(update.args):
                aliases.append((update.relation.fields[i].name, arg))
            return self._make_select(new_lookups, aliases)

        # get a representative update
        update = v.collect_by_type(ir.Update, union).some()

        # TODO - maybe this should be more like INSERT INTO a table than a view?
        self.statements.append(sql.CreateView(update.relation.name,
            sql.CTE(True, f"{update.relation.name}_rec", [f.name for f in update.relation.fields], [
                make_case_select(cast(ir.Logical, union.tasks[0])),
                make_case_select(cast(ir.Logical, union.tasks[1]))
            ]))
        )


    def _create_statement(self, logical: ir.Logical):

        # TODO - improve the typing info to avoid these casts
        lookups = cast(list[ir.Lookup], filter_by_type(logical.body, ir.Lookup))
        updates = cast(list[ir.Update], filter_by_type(logical.body, ir.Update))
        outputs = cast(list[ir.Output], filter_by_type(logical.body, ir.Output))

        # TODO - this is simplifying soooo much :crying_blood:
        if updates and not lookups:
            # TODO: this is assuming that the updates are all static values
            # insert static values: INSERT INTO ... VALUES ...
            for u in updates:
                r = u.relation
                tuples = self._get_tuples(logical, u)
                for tuple in tuples:
                    self.statements.append(
                        sql.Insert(r.name, [f.name for f in r.fields], tuple, None)
                    )
        elif updates and lookups:
            # insert values that match a query: INSERT INTO ... SELECT ... FROM ... WHERE ...
            for u in updates:
                r = u.relation
                aliases = []
                for i, arg in enumerate(u.args):
                    aliases.append((r.fields[i].name, arg))

                self.statements.append(
                    sql.Insert(r.name,
                               [f.name for f in r.fields], [],
                               self._make_select(lookups, aliases)
                    )
                )
        elif outputs and lookups:
            # output a query: SELECT ... FROM ... WHERE ...
            aliases = []
            for o in outputs:
                aliases.extend(o.aliases)
            self.statements.append(self._make_select(lookups, aliases))
            pass
        else:
            raise Exception(f"Cannot create SQL statement for:\n{logical}")



    def _make_select(self, lookups: list[ir.Lookup], outputs: list[Tuple[str, ir.Value]]) -> sql.Select:
        table_lookups = filter(lambda t: not builtins.is_builtin(t.relation), lookups)
        froms = []
        sql_vars: dict[ir.Lookup, str] = dict() # one var per table lookup
        var_column: dict[Tuple[ir.Var, ir.Lookup], ir.Field] = dict()
        var_sql_var: dict[ir.Var, str] = dict()
        var_lookups: dict[ir.Var, OrderedSet[ir.Lookup]] = dict()
        i = 0
        for lookup in table_lookups:
            varname = f"v{i}"
            i += 1
            froms.append(sql.From(lookup.relation.name, varname))
            sql_vars[lookup] = varname
            j = 0
            for arg in lookup.args:
                if isinstance(arg, ir.Var):
                    var_column[arg, lookup] = lookup.relation.fields[j]
                    var_sql_var[arg] = varname
                    if arg not in var_lookups:
                        var_lookups[arg] = ordered_set()
                    var_lookups[arg].add(lookup)
                j += 1

        computed_outputs: dict[ir.Var, str] = dict()

        wheres: list[sql.Expr] = []
        # uses of built-in expressions
        def reference(v):
            if isinstance(v, ir.Var):
                # TODO - assuming the built-in reference was grounded elsewhere
                lookup = var_lookups[v].some()
                return f"{var_sql_var[v]}.{var_column[(v, lookup)].name}"
            else:
                # assuming a literal
                return str(v)
        for lookup in filter(lambda t: builtins.is_builtin(t.relation), lookups):
            # TODO - assuming infix binary or ternary operators here
            lhs = reference(lookup.args[0])
            rhs = reference(lookup.args[1])
            if len(lookup.args) == 3 and isinstance(lookup.args[2], ir.Var):
                # TODO - assuming that the computed value is going to be returned, which is obviously generally wrong
                computed_outputs[lookup.args[2]] = f"{lhs} {lookup.relation.name} {rhs}"
            else:
                wheres.append(sql.Terminal(f"{lhs} {lookup.relation.name} {rhs}"))
        # if there are 2 lookups for the same variable, we need a join
        for arg, lookup_set in var_lookups.items():
            if len(lookup_set) > 1:
                refs = [f"{sql_vars[lu]}.{var_column[cast(ir.Var, lu.args[0]), lu].name}" for lu in lookup_set]
                # join variable references pairwise (e.g. "x.id = y.id AND y.id = z.id")
                for lhs, rhs in zip(refs, refs[1:]):
                    wheres.append(sql.Terminal(f"{lhs} = {rhs}"))

        # conjunction of wheres
        if len(wheres) == 0:
            where = None
        elif len(wheres) == 1:
            where = sql.Where(wheres[0])
        else:
            where = sql.Where(sql.And(wheres))

        # finally, compute what the select will return
        vars = []
        for alias, var in outputs:
            if var in var_lookups:
                lookup = var_lookups[var].some()
                vars.append(sql.VarRef(sql_vars[lookup], var_column[var, lookup].name, alias))
            elif var in computed_outputs:
                # TODO - abusing VarRef.name here, it's actually an expression here. Fix it!
                vars.append(sql.VarRef(computed_outputs[var], None, alias))
            else:
                # TODO - abusing even more here, because var is a value!
                vars.append(sql.VarRef(str(var), None, alias))


        return sql.Select(False, vars, froms, where)



    def _get_tuples(self, logical: ir.Logical, u: ir.Update):
        """
        Get a list of tuples to perform this update.

        This function traverses the update args, assuming they contain only static values or
        variables bound to a construct task, and generates a list of tuples to insert. There
        may be multiple tuples because arguments can be lists of values bound to a field
        whose role is multi.
        """
        # TODO - this only works if the variable is bound to a Construct task, we need a
        # more general approach.
        values = []
        for a in u.args:
            if isinstance(a, ir.Var):
                for t in logical.body:
                    if isinstance(t, ir.Construct) and t.id_var == a:
                        values.append(f"hash({', '.join([self._convert_value(v) for v in t.values])})")
                        break
            elif isinstance(a, FrozenOrderedSet):
                values.append(frozen(*[self._convert_value(v) for v in a]))
            else:
                values.append(self._convert_value(a))
        return self._product(values)

    def _product(self, values):
        """ Compute a cartesian product of values when the value is a FrozenOrderedSet. """
        # TODO - some pass needs to check that this is correct, i.e. that we are using a
        # FrozenOrderedSet only if the field is of role multi.
        tuples = [[]]
        for value in values:
            if isinstance(value, FrozenOrderedSet):
                tuples = [prev + [element] for prev in tuples for element in value]
            else:
                tuples = [prev + [value] for prev in tuples]
        return tuples

    def _convert_value(self, v) -> str:
        """ Convert the literal value in v to a SQL value."""
        if isinstance(v, str):
            return f"'{v}'"
        if isinstance(v, ir.ScalarType):
            return f"'{v.name}'"
        if isinstance(v, ir.Literal):
            return self._convert_value(v.value)
        return v

    BUILTIN_CONVERSION = {
        "Hash": "UBIGINT",
        "String": "TEXT",
        "Number": "DOUBLE",
        "Int": "INT",
        "Decimal": "DECIMAL(13, 2)",
        "Bool": "BOOLEAN"
    }
    def _convert_type(self, t: ir.Type) -> str:
        """ Convert the type t into the equivalent SQL type."""
        if isinstance(t, ir.SetType):
            return self._convert_type(t.element_type)

        # entities become HUGEINT
        if not types.is_builtin(t):
            return "HUGEINT"

        # convert known builtins
        if isinstance(t, ir.ScalarType) and t.name in self.BUILTIN_CONVERSION:
            return self.BUILTIN_CONVERSION[t.name]
        raise Exception(f"Unknown built-in type: {t}")



from relationalai.early_access.dsl import AbstractRelation, RelationSignature
from relationalai.early_access.dsl.bindings.common import BindableAttribute


class AttributeView(AbstractRelation):
    """
    A class representing an attribute view relation.
    """

    def __init__(self, namespace, attr: 'BindableAttribute'):
        name = self._get_relation_name(attr)
        sig = self._get_relation_signature(attr)
        super().__init__(namespace, name, sig)
        self._attr = attr

    @staticmethod
    def _get_relation_name(attr: 'BindableAttribute'):
        """
        Returns the relation name of the attribute view, combining table name and attribute name.
        """
        attr_name = attr.physical_name()
        table_name = attr.table.physical_name()
        name = f"{table_name}:{attr_name}"
        return name

    @staticmethod
    def _get_relation_signature(attr: 'BindableAttribute'):
        """
        Returns the relation signature of the attribute view.
        """
        sig = RelationSignature(attr.table.key_type(), attr.type())
        return sig

    def attr(self):
        """
        Returns the attribute of the attribute view.
        """
        return self._attr

    def __repr__(self):
        return f"AttributeView({self.qualified_name()})"



from relationalai.early_access.dsl.core import warn
from relationalai.early_access.dsl.core.constraints import FalseConstraint
from relationalai.early_access.dsl.core.constraints.predicate.atomic import reduce_relational_atoms
from relationalai.early_access.dsl.core.constraints.scalar import entails, fold_constants, reduce_constraints, unsatisfiable
from relationalai.early_access.dsl.core.exprs.scalar import Literal, box_number
from relationalai.early_access.dsl.core.logic import LogicFragment
from relationalai.early_access.dsl.core.logic.helper import occurrences, instantiate_type_constraints_for
from relationalai.early_access.dsl.core.logic.exists import existentially_quantify
from relationalai.early_access.dsl.rulesets import RuleSet

class SchemaFragment(LogicFragment):

    def __init__(self):
        super().__init__()

        self._relations = {}
        self._universals = {}

    # Attempts to add relational declaration d to the declarations part of this
    # SchemaFragment. If d is new to this fragment, returns True. If the variable
    # d declares already exists in this fragment and is type consistent with d,
    # then returns False. If the variable exists and is type inconsistent with d,
    # raises an exception.
    #
    def add_relational_component(self, d):
        vname = d.display()
        if vname not in self._relations:
            self._relations[vname] = d
            return True
        else:
            d1 = d.typeof()
            d2 = self._relations[vname].typeof()
            if not d2.subtype_of(d1):
                if d1.subtype_of(d2):
                    self._relations[vname] = d
                else:
                    raise Exception(
                        f"++ Error: Cannot declare two relational variables named {vname} in the same schema context")

    # Add c to the constraint part of this Schema. Note that we distinguish propositional
    # from predicate constraints based on the nature of c. Propositional constraints are
    # used directly in the bodies of generated rules and queries. Predciate constraints
    # are used to reason about how to generate rules that populate relations.
    #
    def add_predicate_constraint(self, c):
        cv = c.entityid()
        if not c.relational():
            raise Exception(f"Attempt to add {c.pprint()} as a predicate constraint")

        if cv not in self._universals:
            self._universals[cv] = c

    # Let self be an operation schema fragment and pre an undecorated state schema
    # fragment. This operation returns a fragment that is self reduced by constraints
    # that are entailed by what remains if we assume pre.
    #
    def assume_precondition(self, pre):
        tconsts = self._sconstraints

        tlen = len(tconsts) + 1
        while(len(tconsts) < tlen):
            tlen = len(tconsts)

            for c in tconsts:
                # We can remove c from self if c is entailed by:
                #
                #    context =^= (self \ {c}) u pre
                #
                tcprime = { k: tconsts[k] for k in tconsts if k != c }
                context = tcprime | pre._sconstraints

                if entails(context, tconsts[c]):
                    tconsts = tcprime
                    break

        new_trans = self.duplicate()
        new_trans._sconstraints = tconsts
        return new_trans

    # Returns the schema formed by concealing under an existential quantifier any component
    # variables whose names appear in the list cvars.
    #
    # Note: Clients typically call this method indirectly from the hide method, where they
    #       pass in the components directly using named Schema instance variables.
    #
    def conceal(self, cvars):

        edecls = {c: self._scalars[c] for c in cvars}

        s = SchemaFragment()
        s._relations = self._relations
        s._scalars = {v: self._scalars[v] for v in self._scalars if v not in edecls}
        s._universals = self._universals

        (s._sconstraints,
         s._atoms,
         xc) = existentially_quantify(edecls, self._sconstraints, self._atoms)

        s.add_existential_constraint(xc)
        return s

    # Computes a dictionary of constraints that are declared in rcdict
    # but not in lcdict after first renaming the variables of rcdict so
    # that its constraints will refer to the same exact SchemaComponents
    # as are used in self for any variables that are common.
    #
    # We use this extension when conjoining these two Schemas.
    #
    def constraint_extension(self, lcdict, rcdict):
        ext = {}
        decls = dict(self._scalars, **self._relations)
        for c in rcdict.values():
            cv = c.entityid()
            if cv not in lcdict:
                ext[cv] = c.rename(decls)
        return ext

    # Returns a decorated (dashed) version of this SchemaFragment
    #
    def decorate(self):
        scalars = self._scalars
        relations = self._relations
        scalar_renaming = { v : scalars[v].after for v in scalars }
        rel_renaming = { v : relations[v].after for v in relations }
        return self.rewrite(scalar_renaming, rel_renaming)

    # Given a tuple of relational vars, return a new SchemaFragment that is a duplicate
    # of self with RelationIsEmpty constraints on all v in vars
    #
    def is_empty(self, *vars):
        s = self.rewrite(self._scalars, self._relations)
        for v in vars:
            s.add_predicate_constraint(v.empty())
        return s

    # Returns a prefixed version of this SchemaFragment in which every
    # component is prefixed by pref (a string)
    #
    def prefix(self, pref):
        scalars = self._scalars
        relations = self._relations
        scalar_renaming = { v : scalars[v].prefix(pref) for v in scalars }
        rel_renaming = { v : relations[v].prefix(pref) for v in relations }

        return self.rewrite(scalar_renaming, rel_renaming)

    def rewrite(self, scalar_renaming, rel_renaming):
        newfrag = SchemaFragment()

        for v in self._scalars:
            if v in scalar_renaming:
                newvar = self._scalars[v].rename(scalar_renaming)
                newfrag._scalars[newvar.display()] = newvar
            else:
                newfrag._scalars[v] = self._scalars[v]

        for v in self._relations:
            if v in rel_renaming:
                newvar = self._relations[v].rename(rel_renaming)
                newfrag._relations[newvar.display()] = newvar
            else:
                newfrag._relations[v] = self._relations[v]

        scalarcs = {}
        for c in self._sconstraints.values():
            newc = c.rename(scalar_renaming)
            scalarcs[newc.entityid()] = newc
        newfrag._sconstraints = scalarcs

        relcs = {}
        for c in self._atoms.values():
            newc = c.rename(rel_renaming).rename(scalar_renaming)
            relcs[newc.entityid()] = newc
        newfrag._atoms = relcs

        # [REKS TODO] Verify that we properly rename the relation components
        #             in existentials.
        #
        xcs = {}
        for c in self._existentials.values():
            newc = c.rename(scalar_renaming).rename(rel_renaming)
            xcs[newc.entityid()] = newc
        newfrag._existentials = xcs

        predcs = {}
        for c in self._universals.values():
            newc = c.rename(rel_renaming)
            predcs[newc.entityid()] = newc
        newfrag._universals = predcs

        return newfrag

    # Return the list of decorated positively bound atoms from among the propositional
    # constraints of this Schema. We use these to extract Rel rule(s) from the Schema
    #
    def decorated_atoms(self):
        return {c.entityid(): c
                for c in self._atoms.values()
                if c.decorated()}

    def duplicate(self):
        newFragment = SchemaFragment()

        scalars = self._scalars
        relations = self._relations
        newFragment._scalars = {v: scalars[v] for v in scalars}
        newFragment._relations = {v: relations[v] for v in relations}
        scalarcs = self._sconstraints
        relcs = self._atoms
        predcs = self._universals
        xcs = self._existentials
        newFragment._sconstraints = {c: scalarcs[c] for c in scalarcs}
        newFragment._atoms = {c: relcs[c] for c in relcs}
        newFragment._existentials = { c: xcs[c].duplicate() for c in xcs }
        newFragment._universals = {c: predcs[c] for c in predcs}

        return newFragment

    # Partially evaluate this Schema by substituting values for
    # its schema variables, simplifying where possible.
    #
    def evaluate(self, bdgs):

        decls = {}
        for v in self._scalars:

            var = self._scalars[v]
            if v in bdgs:
                # Then type check val against var
                var_type = var.typeof()
                val = bdgs[v]

                if not var_type.contains(val.val if isinstance(val, Literal) else val):
                    return None

                if not isinstance(val, Literal):
                    # ... val needs to be boxed before checking other constraints
                    bdgs[v] = box_number(val)
            else:
                decls[v] = self._scalars[v]

        decls = { v: self._scalars[v] for v in self._scalars if v not in bdgs}

        super = LogicFragment.evaluate(self, bdgs)

        if super is None:
            return super
        else:
            s = SchemaFragment()
            s._scalars = decls
            s._sconstraints = super._sconstraints
            s._atoms = super._atoms
            s._relations = self._relations
            s._universals = self._universals
            return s

    # Merge this SchemaFragment with some other SchemaFragment to produce a new one
    # that is the conjunction of the two.
    #
    def merge(self, other):
        s = SchemaFragment()
        s._scalars = merge_decls(self._scalars, other._scalars)
        s._relations = merge_decls(self._relations, other._relations)
        #
        propc = self._sconstraints.copy()
        propc.update(self.constraint_extension(self._sconstraints, other._sconstraints))
        s._sconstraints = propc
        #
        propc = self._atoms.copy()
        propc.update(self.constraint_extension(self._atoms, other._atoms))
        s._atoms = propc
        #
        predc = self._universals.copy()
        predc.update(self.constraint_extension(self._universals, other._universals))
        s._universals = predc
        #
        return s

    # Returns True if this SchemaFragment is an OperationFragment -- meaning
    # that each component is either decorated or undecorated and there exists
    # at least one decorated AtomicConstraint
    #
    def operation(self):
        isop = False
        for c in self._scalars.values():
            if not c.decorated() and not c.undecorated():
                return False

        for c in self._relations.values():
            if not c.decorated() and not c.undecorated():
                return False

        for c in self._atoms.values():
            if c.decorated():
                isop = True
            else:
                if not c.undecorated():
                    return False

        return isop

    # Emit this schema to a textual output using the Z display style
    #
    def pprint(self):
        # Separator used when pretty-printing Schemas in the Z style
        componentSeparator = "+----------"

        declStrings = []
        declStrings.append(componentSeparator)

        for v in self._relations:
            declStrings.append(f"| {self._relations[v].pprint()}")

        for v in self._scalars:
            declStrings.append(f"| {self._scalars[v].pprint()}")

        declStrings.append(componentSeparator)

        propcs = self.propositional_constraints()

        if len(propcs) + len(self._universals) > 0:
            predStrings = [f"| {c.pprint()}      # [ic]" for c in self._universals.values()]
            propStrings = [f"| {c.pprint()}" for c in propcs.values()]

            declStrings.extend(predStrings)
            declStrings.extend(propStrings)
            declStrings.append(componentSeparator)

        return "\n".join(declStrings)

    # A Schema is propositional if none of its variables is of relation type.
    #
    def propositional(self):
        return len(self._relations) == 0

    def reduce(self):
        s = self.reduce_relational_atoms()
        if s is None:
            return s
        else:
            return s.reduce_scalar_constraints()

    # Simplify the scalar propositional constraints of this Schema using
    # satisfiability checks over numeric inequalities and expressions.
    #
    # If we can decide a Schema is unstaisfiable, return None
    #
    def reduce_scalar_constraints(self):
        C = self._sconstraints.copy()

        atoms = self._atoms
        valueTypeConstraints = infer_referent_constraints(atoms)
        for v in valueTypeConstraints:
            C[v] = valueTypeConstraints[v]

        if unsatisfiable(C):
            return None

        reduced = reduce_constraints(C)
        newprops = {c: reduced[c] for c in reduced if c not in valueTypeConstraints}

        s = SchemaFragment()
        s._scalars = self._scalars
        s._relations = self._relations
        s._universals = self._universals
        s._atoms = self._atoms
        s._sconstraints = newprops

        return s

    # Simplify thus Schema by projecting out decls that are equated with other
    # decls via predicate constraints. These commonly arise when decorating
    # Schemas using the Xi operator. When reducing we try to remove the SchemaComponent
    # with the larger number of dashes to simplify Rel rule generation.
    #
    def reduce_by_xi(self):
        simplifyingICs = [c.xiSubstitution() for c in self._universals.values() if c.xi()]
        if len(simplifyingICs) > 0:
            bindings = {c.left.display(): c.right for c in simplifyingICs}

            s = SchemaFragment()
            s._scalars = self._scalars
            s._relations = self._relations # todo: verify the implementation for projectDecls(self._relations, bindings)
            s._sconstraints = self._sconstraints

            for c in self._atoms.values():
                cp = c.rename(bindings)
                s._atoms[cp.entityid()] = cp

            for c in self._universals.values():
                if not c.xi():
                    cp = c.rename(bindings)
                    s._universals[cp.entityid()] = cp

            return s
        else:
            return self

    # Reduce the AtomicConstraints of this Schema using unification, where
    # possible, and existential reduction.
    #
    # Returns None if FalseConstraints are detected during reduction.
    #
    def reduce_relational_atoms(self):
        (neweqs, reduced_atoms) = reduce_relational_atoms(self._atoms)
        sconsts = neweqs | self._sconstraints
        (newcomps, newatoms) = fold_constants(sconsts, reduced_atoms)
        if len(newcomps) == 1:
            for c in newcomps.values():
                if isinstance(c, FalseConstraint):
                    return None

        s = SchemaFragment()
        s._universals = self._universals
        s._sconstraints = newcomps
        s._atoms = newatoms
        s._scalars = self._scalars
        s._relations = self._relations
        return s

    # Attempt to refine this OperationFragment into a RuleSet that implements it
    #
    def refine(self):
        if not self.operation():
            raise Exception(f"Can only refine OperationFragments, but this isn't one:\n {self.pprint()}\n----")

        datoms = self.decorated_atoms()
        satoms = self._atoms
        body_atoms = {c: satoms[c] for c in satoms if c not in datoms}

        rs = RuleSet(datoms, self._sconstraints, body_atoms)

        grounding_constraints = {} | self._sconstraints | body_atoms

        c = self.ungrounded_components(grounding_constraints)
        if len(c) > 0:
            # Ungrounded variables could signal a spec that is too loose to fully refine
            # into datalog. If the vars in c occur in at least one constraint, then we'll
            # need to warn the user and mark the generated RuleSet as inline
            #
            ungrounded_occurrences = []
            for v in c:
                vname = v.display()
                if occurrences(vname, grounding_constraints) != 0 and occurrences(vname, datoms) != 0:
                    ungrounded_occurrences.append(v)
            if len(ungrounded_occurrences) > 0:
                vstr = ", ".join([v.display() for v in c])
                warn(f"Ungrounded variables ({vstr}) in schema prevent materialization (declaring view @inline)")
                rs.make_inline()

        return rs

    def rename(self, xform):

        scalars = self._scalars
        relations = self._relations
        scalar_renaming = {}
        rel_renaming = {}

        for v in xform:
            if v in scalars:
                scalar_renaming[v] = xform[v]
            else:
                if v in relations:
                    rel_renaming[v] = xform[v]

        return self.rewrite(scalar_renaming, rel_renaming)


# Given a set A of AtomicConstraints, look for any variables used in a
# role that is played by a ValueType that applies constraints and
# return the constraints instantiated by the variable(s)
#
def infer_referent_constraints(A):
    constraints = {}
    # [REKS] Go fill this in to actually use constraints that we can
    #        infer from positive AtomicConstraint references
    #
    for a in A.values():
        role_players = a.relcomp.typeof()._types
        args = a.args
        for i in range(len(role_players)):
            tp = role_players[i]
            expr = args[i]
            if tp.constraint_unary() and expr.variable():
                tcs = instantiate_type_constraints_for(expr, tp)
                constraints.update(tcs)
    return constraints


# Merges the declarations of this Schema and another to produce a SchemaComponent
# dictionary that is suitable for a Scheme formed from the composition of these
# two Schemas.
#
# Returns a SchemaComponent dictionary (decls) that satisfies the following:
#  - decls == self.decls UNION (other.decls NDRES dom(self.decls))
#  - for all v in dom(self.decls) INTERSECT dom(other.decls)
#        self.decls[v].typeof() == other.decls[v].typeof()
# Raises an exception if Schemas declare the same variable with incompatible
# types.
#
def merge_decls(left, right):
    decls = dict(left)

    for v in right:
        svar = right[v]
        if v in decls:
            if decls[v].typeof().entityid() != svar.typeof().entityid():
                raise Exception(f"++ During schema composition, operands declare incompatible types for variable {v}.")
        else:
            decls[v] = svar

    return decls



from dataclasses import dataclass
from typing import Optional


@dataclass
class ORMEntityType:
    id: str
    name: str
    preferred_id: Optional[str]

@dataclass
class ORMValueType:
    id: str
    name: str
    data_type: str

@dataclass
class ORMRole:
    id: str
    name: str
    relationship_name: str
    player: str

@dataclass
class SubtypeArrow:
    start: Optional[str]
    end: Optional[str]

@dataclass
class ORMUniquenessConstraint:
    id: str
    roles: list[str]
    identifies: Optional[str]